

__all__ = ['IO']


from .utils.io import *


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='Src - Ref', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input spectra temperature or flux file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')

parser.add_argument('--t_src', type=float, required=True,
                   help='on-source time[second]')
parser.add_argument('--t_ref', type=float, required=True,
                   help='off-source time[second]')
parser.add_argument('--t_change', type=float, required=True, choices=[30,60],
                    help="switching time[second]. 30s for sepatation of src and ref less than 20'; 60s for sepatation between 20' and 60'.")
parser.add_argument('--n_repeat', type=int, required=True,
                    help='The number of on-source off-source cycles')

parser.add_argument('--only_off', type=bool_fun, choices=[True, False], default='False',
                    help='only noise off')
parser.add_argument('--only_src', type=bool_fun, choices=[True, False], default='False',
                    help='no ref')


class IO(BaseIO):
    ver = 'old'

    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-ps'
        return fpart

    def _import_m(self,):
        super()._import_m()
        global np
        import numpy as np

    def _prepare(self, *args, **kwargs):
        # freq: self.freq_use
        self.mjd_a = self.mjd
        # noise-diode on
        is_on = self.fs['is_on'][:]

        self.inds_on = np.where(is_on)[0]
        self.inds_off = np.where(~is_on)[0]
        self.mjd_on = self.mjd_a[self.inds_on]
        # noise-diode off
        self.mjd_off = self.mjd_a[self.inds_off]
        # load T
        self.p_on = self.s2p[self.inds_on]
        self.p_off = self.s2p[self.inds_off]

#         self.Tcal_s = self.get_Tcal_s()


    @staticmethod
    def in_range(arr, starts, length):
        """
        arr: shape (m,)
        starts: shape (n,)
        length: scalar
        """
        is_in =  (arr[:,None] > starts[None,:]) & (arr[:,None] < (starts+length)[None,:])
        return np.logical_or.reduce(is_in, axis=1)

    def sep(self, t_src, t_ref, n_repeat, t_change=30):
        """
        second
        """
        #######################
        # on: noise-diode on
        # off: noise-diode off
        # src: on-source
        # ref: off-source
        #######################

        # second to day
        t_src = t_src/60/60/24
        t_ref = t_ref/60/60/24
        n_repeat = n_repeat
        t_change = t_change/60/60/24

        mjd_src_start = self.mjd_a[0] + np.arange(n_repeat)*(t_src + t_change + t_ref + t_change)
        mjd_ref_start = mjd_src_start + (t_src + t_change)

        is_ = self.in_range(self.mjd_on, mjd_src_start, t_src)
        self.p_on_src = self.p_on[is_]
        self.inds_on_src = self.inds_on[is_]

        is_ = self.in_range(self.mjd_off, mjd_src_start, t_src)
        self.p_off_src = self.p_off[is_]
        self.inds_off_src = self.inds_off[is_]

        is_ = self.in_range(self.mjd_on, mjd_ref_start, t_ref)
        self.p_on_ref = self.p_on[is_]
        self.inds_on_ref = self.inds_on[is_]

        is_ = self.in_range(self.mjd_off, mjd_ref_start, t_ref)
        self.p_off_ref = self.p_off[is_]
        self.inds_off_ref = self.inds_off[is_]

        self.mjd_src_start = mjd_src_start
        self.mjd_ref_start = mjd_ref_start
        
        
    def gen_integrated_time(self, ):
        args = self.args
        
        tint_src = np.sum(~np.all(np.isnan(self.p_off_src[:,:,-1]), axis = 1))
        tint_ref = np.sum(~np.all(np.isnan(self.p_off_ref[:,:,-1]), axis = 1))
        
        if not args.only_off:
            tint_src += np.sum(~np.all(np.isnan(self.p_on_src[:,:,-1]), axis = 1))
            tint_ref += np.sum(~np.all(np.isnan(self.p_on_ref[:,:,-1]), axis = 1))
        
        from astropy.time import Time
        mjd = Time(self.mjd, format='mjd')
        delta_mjd = (mjd[1] - mjd[0]).sec
        
        self.tint_src = tint_src * delta_mjd
        self.tint_ref = tint_ref * delta_mjd

    def gen_Ta(self, only_off=False):
        """
        using

        self.p_on_src
        self.p_off_src
        self.p_on_ref
        self.p_off_ref

        to gen

        self.Ta as shape (1,chan,polar)

        """
        args = self.args

        Ta_on_src = np.nanmean(self.p_on_src, axis=0, keepdims=True)
        Ta_off_src = np.nanmean(self.p_off_src, axis=0, keepdims=True)
        Ta_on_ref = np.nanmean(self.p_on_ref, axis=0, keepdims=True)
        Ta_off_ref = np.nanmean(self.p_off_ref, axis=0, keepdims=True)

        # use interaged time as weight
        if not args.only_off:
            Ta_src = np.average(np.r_[Ta_on_src, Ta_off_src], axis=0, weights=[len(self.inds_on_src),len(self.inds_off_src)])
            Ta_ref = np.average(np.r_[Ta_on_ref, Ta_off_ref], axis=0, weights=[len(self.inds_on_ref),len(self.inds_off_ref)])
        else:
            Ta_src = Ta_off_src[0]
            Ta_ref = Ta_off_ref[0]

        self.Ta_src = Ta_src[None,]
        self.Ta_ref = Ta_ref[None,]

        if args.only_src:
            self.Ta = self.Ta_src
        else:
            self.Ta = self.Ta_src - self.Ta_ref

        # also change mjd
        inds_tmp = self.inds_off_src[:1]
        self.mjd = self.mjd_a[inds_tmp]

    def gen_radec(self,):
        """
        try to calculation for each record: self.ra_a, self.dec_a
        use the position of first src record as self.ra, self.dec
        also set self.mjd and self.mjd_a

        """

        self.ra_a, self.dec_a = self.ra, self.dec
        inds_tmp = self.inds_off_src[:1]
        self.ra, self.dec = self.ra_a[inds_tmp], self.dec_a[inds_tmp]
        self.mjd = self.mjd_a[inds_tmp]

    def plot_radec(self, figsize=[10,10], outname=None):
        """
        plot the ra dec to check the separation of src and ref
        """

        from matplotlib import pyplot as plt
        plt.figure(figsize=figsize)
        plt.scatter(self.ra_a[self.inds_on_src], self.dec_a[self.inds_on_src], s=1, color='r')
        plt.scatter(self.ra_a[self.inds_off_src], self.dec_a[self.inds_off_src], s=1, color='r', label='src')
        plt.scatter(self.ra_a[self.inds_on_ref], self.dec_a[self.inds_on_ref], s=1, color='b')
        plt.scatter(self.ra_a[self.inds_off_ref], self.dec_a[self.inds_off_ref], s=1, color='b', label='ref')
        plt.xlabel('ra')
        plt.ylabel('dec')
        plt.grid()
        plt.minorticks_on()
        plt.legend()
        if outname is not None:
            print(f'Saving ra dec plot to {outname}')
            plt.savefig(outname)

    def gen_s2p_out(self,):
        args = self.args
        self._prepare()
        self.sep(args.t_src, args.t_ref, args.n_repeat, args.t_change)
        self.gen_integrated_time()
        self.gen_Ta()
        self.gen_radec()
        self.plot_radec(outname=self.fpath_out + '-radec.png')
        self.s2p_out = self.Ta

    def __call__(self, save=True):
        # store output data in self.dict_out
        self.gen_s2p_out()
        self.gen_dict_out()
        for key in ['is_delay', 'is_on', 'next_to_cal']:
            if key in self.dict_out.keys():
                 self.dict_out[key] = np.array([False,])
        self.dict_out['Ta_src'] = MjdChanPolar_to_PolarMjdChan(self.Ta_src)
        self.dict_out['Ta_ref'] = MjdChanPolar_to_PolarMjdChan(self.Ta_ref)
        
        
        for key in ['tint_src', 'tint_ref']:
            print(key, getattr(self, key))
            self.dict_out[key] = np.array([getattr(self, key)])
        
        # save to hdf5 file
        if save:
            self.save()


if __name__ == '__main__':
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()



## cal
"""
Tcal Data Management Module
---------------------------

This module handles reading and automatically updating Tcal (temperature calibration) data.

Key Features & Logic:
1. **GitHub-based Storage**: Data is hosted on GitHub Releases. A `manifest.json` lists available dates.
   - Repo: https://github.com/jyingjie/hifast-tcal-data

2. **On-Demand Downloading (Lazy Load)**:
   - Does NOT download all historical data by default.
   - `check_and_update_tcal` fetches the `manifest.json` (cached locally for 24h).
   - Only when a specific date is needed (and missing locally) will it be downloaded.

3. **Smart 'Auto' Selection**:
   - `read_tcal(..., date='auto')` calculates the best match using the UNION of Local files and Remote manifest dates.
   - This ensures the best calibration date is selected even if the file hasn't been downloaded yet.

4. **Concurrency Safety**:
   - Uses `fcntl` non-blocking file locks.
   - Safe to run with `hifast ... -p 50` (multiple parallel processes).
   - Only one process performs the download; others skip or wait.

5. **Offline Mode**:
   - Set environment variable `HIFAST_OFFLINE=1` to disable all network requests.
   - In offline mode, logic degrades gracefully to use ONLY locally available files.
"""
import numpy as np
import os

def read_tcal_sav(nB, s_type='w', tcal_dir=None, mode='high', date='20190115'):
    """
    nB: int
      beam number
    """
    from scipy.io.idl import readsav
    import os
    if tcal_dir is None:
        tcal_dir = _get_default_tcal_dir()
    fname = os.path.join(tcal_dir, f'{date}/median_{date}.Tcal-results.HI_{s_type}.{mode}.sav')
    tc_info= readsav(fname)[f'{mode}_{s_type}'][0]
    tc_freq= tc_info['freq']
    return tc_freq, tc_info['M%02d_TC'%nB], fname

def read_tcal_fits(nB, s_type='w', tcal_dir=None, mode='high', date=''):
    """
    nB: int
      beam number
    """
    from astropy.io import fits
    import os
    if tcal_dir is None:
        tcal_dir = _get_default_tcal_dir()
    fname = os.path.join(tcal_dir, f'{date}/CAL.{date}.{mode}.{s_type.upper()}.fits')
    f = fits.open(fname)
    tc_freq = f[1].data['FREQ'][0]
    tc_T = f[1].data['TCAL'][0,nB-1].T
    return tc_freq, tc_T, fname




from .config import conf

def _get_default_tcal_dir():
    raw_dir = conf.get('tcal', 'tcal_dir', '~/Tcal/')
    return os.path.expanduser(raw_dir)

# URLs resolved via config system (Env > Config File > Default)
TCAL_REPO_MANIFEST_URL, TCAL_REPO_BASE_URL = conf.get_tcal_urls()
TCAL_UPDATE_INTERVAL = 86400  # Check once per 24 hours

def check_and_update_tcal(tcal_dir, target_date=None):
    """
    Checks for Tcal updates, caches the manifest, and optionally downloads a specific date.
    
    Args:
        tcal_dir (str): Local Tcal directory.
        target_date (str, optional): Specific date string (e.g., '20200531') to ensure is available.
        
    Returns:
        list: List of available date strings (remote manifest).
    """
    import time
    import json
    import fcntl
    import requests
    import os  # Added import os
    from .downloader import download_and_extract_zip

    # 1. Offline Mode Check
    if conf.get_boolean('general', 'offline'):
        return []

    os.makedirs(tcal_dir, exist_ok=True)
    
    timestamp_file = os.path.join(tcal_dir, '.last_update_check')
    manifest_cache_file = os.path.join(tcal_dir, 'manifest_cache.json')
    lock_file_path = os.path.join(tcal_dir, '.update.lock')
    # Phase 1: Separate lock for manifest to avoid race conditions
    manifest_lock_path = os.path.join(tcal_dir, '.manifest.lock') 
    failure_marker = os.path.join(tcal_dir, '.network_failure')
    
    # --- Phase 1: Ensure Manifest Cache is Up-to-Date ---
    # Double-checked locking to prevent redundant downloads
    
    # 1. Optimistic Check
    need_update = True
    if os.path.exists(timestamp_file) and os.path.exists(manifest_cache_file):
        try:
            mtime = os.path.getmtime(timestamp_file)
            if time.time() - mtime < TCAL_UPDATE_INTERVAL:
                need_update = False
        except OSError:
            pass

    if need_update:
        # 2. Acquire Lock and Re-check
        # We use a blocking lock here. Usage of 'a+' matches other lock implementations in this module.
        m_lock = open(manifest_lock_path, 'a+')
        try:
            fcntl.lockf(m_lock, fcntl.LOCK_EX) # Blocking
            
            # Re-check timestamp (someone else might have updated it while we waited)
            already_updated = False
            if os.path.exists(timestamp_file) and os.path.exists(manifest_cache_file):
                try:
                    mtime = os.path.getmtime(timestamp_file)
                    if time.time() - mtime < TCAL_UPDATE_INTERVAL:
                        already_updated = True
                except OSError:
                    pass
            
            if not already_updated:
                try:
                    from .downloader import get_file
                    # Use get_file with overwrite=True to force update
                    res = get_file(TCAL_REPO_MANIFEST_URL, manifest_cache_file, overwrite=True, failure_marker=failure_marker)
                    if res:
                        # Update timestamp
                        with open(timestamp_file, 'w') as f:
                            f.write(str(time.time()))
                except Exception as e:
                    print(f"Warning: Tcal manifest update failed: {e}")
            
        finally:
            fcntl.lockf(m_lock, fcntl.LOCK_UN)
            m_lock.close()

    # --- Phase 2: Load Manifest ---
    manifest_dates = []
    if os.path.exists(manifest_cache_file):
        try:
            with open(manifest_cache_file, 'r') as f:
                data = json.load(f)
                manifest_dates = data.get('date', [])
        except:
             pass

    # --- Phase 3: Download Target Date ---
    if target_date and target_date in manifest_dates:
        target_path = os.path.join(tcal_dir, target_date)
        
        if not os.path.exists(target_path):
            
            # Check for breaker before attempting lock (optimization)
            if os.path.exists(failure_marker):
                 try:
                     if time.time() - os.path.getmtime(failure_marker) < 60:
                          return manifest_dates
                 except: pass

            # If target file is missing, we MUST ensure it gets downloaded.
            # We try to get an exclusive lock.
            lock_file = open(lock_file_path, 'a+') 
            try:
                # First try non-blocking (fast path if we are the only one)
                try:
                    fcntl.lockf(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    # Someone else is holding the lock.
                    # Since we NEED this file and it's missing, we must WAIT.
                    print(f"Waiting for another process to download Tcal data for {target_date}...")
                    fcntl.lockf(lock_file, fcntl.LOCK_EX) # Blocking wait
                
                # Once we have the lock, check existance again (maybe previous owner downloaded it)
                if not os.path.exists(target_path):
                    file_url = f"{TCAL_REPO_BASE_URL}/{target_date}/{target_date}.zip"
                    print(f"Downloading Tcal data for {target_date}...")
                    # Use large timeout (60 retries * 10s = 600s = 10 mins) for slow networks
                    download_and_extract_zip(file_url, target_path, max_retries=60, retry_delay=10, failure_marker=failure_marker)
            
            except Exception as e:
                print(f"Error updating tcal data: {e}")
            finally:
                 try:
                    fcntl.lockf(lock_file, fcntl.LOCK_UN)
                    lock_file.close()
                 except:
                    pass
                 
    return manifest_dates

def read_tcal(nB, s_type='w', tcal_dir=None, mode='high', date='auto', mjd=None):
    """
    nB: int
       beam number
    s_type: str
       type, w or n
    tcal_dir: str
       if not set, use ~/Tcal/
    mode: str
       high or low
    date: str
       example: 20190115 or 20200531, default is None
    mjd:
       if date is 'auto', using the nearest date of tcal
    """
    from glob import glob
    import os
    if tcal_dir is None:
        tcal_dir = _get_default_tcal_dir()
    
    
    # 1. Update manifest and get list of remote dates
    remote_dates = check_and_update_tcal(tcal_dir) # First pass: just cache manifest

    # 2. Get local dates
    local_dates = [os.path.basename(i) for i in glob(tcal_dir + '/20[0-9][0-9][0-9][0-9][0-9][0-9]')]
    
    # 3. Combine for 'auto' selection
    all_dates = sorted(list(set(remote_dates + local_dates)))

    if len(all_dates) == 0:
         raise(ValueError('can not find tcal file. If this is the first run, ensure internet connection or manually place files in ~/Tcal/'))

    if date == 'auto':
        from astropy.time import Time
        if mjd is None:
            raise(ValueError('need input mjd if date is auto'))
        mjds_h = Time([f'{s[:4]}-{s[4:6]}-{s[6:8]} 00:00:00.000' for s in all_dates], format='iso').mjd
        #Select best date
        date = all_dates[np.argmin(abs(mjds_h - mjd))]
    
    # 4. Trigger download if needed (for both explicit and auto date)
    check_and_update_tcal(tcal_dir, target_date=date)

    # 5. Final check (in case download failed or it's just invalid)
    # We check local glob again ensures we only proceed if file is really there
    # (Optional: optimization to check os.path.exists direct)
    dates_have_now = [os.path.basename(i) for i in glob(tcal_dir + '/20[0-9][0-9][0-9][0-9][0-9][0-9]')]
    
    if date not in dates_have_now:
         raise(ValueError(f'can not find tcal file in {date} (Download may have failed or date invalid)'))
         
    ## read tcal
    if date == '20190115':
        return read_tcal_sav(nB, s_type, tcal_dir, mode, date='20190115')
    else:
        return read_tcal_fits(nB, s_type, tcal_dir, mode, date=date)

def list_tcal_dates():
    """
    List available Tcal dates with robust fallback:
    1. Online Manifest (Force Update)
    2. Local Cache (if online fails)
    3. Local Directories (if cache fails)
    """
    import sys
    import glob
    import json # Added import json
    tcal_dir = _get_default_tcal_dir()
    
    print(f"Checking Tcal dates in: {tcal_dir}", file=sys.stderr)
    
    dates = []
    source = "online"
    
    # 1. Try Online/Cache via check_and_update_tcal
    # We cheat a bit: passing a non-existent date 'LIST' forces it to load manifest
    # but skip download. However, check_and_update_tcal doesn't force update manifest
    # unless TTL expired or offline.
    # So we manually force update the manifest first.
    
    manifest_url, _ = conf.get_tcal_urls()
    manifest_cache_file = os.path.join(tcal_dir, 'manifest_cache.json')
    failure_marker = os.path.join(tcal_dir, '.network_failure')
    
    try:
        from .downloader import get_file
        if not conf.get_boolean('general', 'offline'):
             res = get_file(manifest_url, manifest_cache_file, overwrite=True, failure_marker=failure_marker)
             if res:
                 source = f"Manifest ({manifest_url})"
             else:
                 source = "Local Cache (Network Failed, using last known good)"
        else:
             print("Warning: Offline mode enabled. List may be incomplete (network unavailable).", file=sys.stderr)
             source = "Local Cache (Offline Mode)"
             
        # Load Cache
        if os.path.exists(manifest_cache_file):
            with open(manifest_cache_file, 'r') as f:
                data = json.load(f)
                dates = data.get('date', [])
                dates.sort()
        else:
            raise FileNotFoundError("Manifest cache not found")

    except Exception as e:
        print(f"Warning: Failed to fetch/read manifest ({e}).", file=sys.stderr)
        source = "local_scan"
        
        # 3. Fallback: Scan Directories
        print("Warning: Falling back to local directory scan.", file=sys.stderr)
        if os.path.exists(tcal_dir):
            # Find directories that look like YYYYMMDD
            dirs = [d for d in os.listdir(tcal_dir) if os.path.isdir(os.path.join(tcal_dir, d)) and d.isdigit() and len(d) == 8]
            dates = sorted(dirs)
    
    if source == "local_scan":
        print(f"Source: Local Scan (Potentially incomplete)", file=sys.stderr)
    else:
        print(f"Source: {source}", file=sys.stderr)
        
    for d in dates:
        print(d)

def download_all_tcal():
    """Downloads all dates available in the manifest."""
    tcal_dir = _get_default_tcal_dir()
    dates = check_and_update_tcal(tcal_dir) # Get manifest
    
    print(f"Found {len(dates)} dates. Starting batch update...")
    for date in dates:
        print(f"Checking/Downloading {date}...")
        check_and_update_tcal(tcal_dir, target_date=date)
    print("All dates processed.")

def get_parser():
    import argparse
    parser = argparse.ArgumentParser(description="Manage Tcal Data")
    subparsers = parser.add_subparsers(dest='command')
    
    # list command
    subparsers.add_parser('list', help='List available Tcal dates (Online > Cache > Local)')
    
    # update-all command
    subparsers.add_parser('update-all', help='Download/Update ALL dates from manifest')
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    
    if args.command == 'list':
        list_tcal_dates()
    elif args.command == 'update-all':
        download_all_tcal()
    else:
        parser.print_help()

if __name__ == "__main__":
    main()





__all__ = []


from .utils.io import *
from .cube import parser
import copy


parser.add_argument('--share_mem_dir', default='/dev/shm',
                   help='Default is `/dev/shm` which is in memory.')
hide_paras(parser, 'share_mem')
for act in parser._actions:
    if act.dest == 'method':
        act.choices += ['mean', 'median', 'wmean', 'wmedian']
        break


if __name__ == '__main__':
    import os
    import tempfile
    from .core.image2 import Imaging2
    args_ = parser.parse_args()
    args_.share_mem = True
    tempfile.tempdir = args_.share_mem_dir

    img2 = Imaging2(args_)
    img2()





__all__ = ['SpecFile', 'align_chan', 'Imaging', 'CalcSpecPixel', 'args']


import numpy as np
import h5py
from glob import glob
import copy
import os
import sys
import copy

from astropy.coordinates import SkyCoord
from astropy import units as u

from .wcs import gen_header, gen_grid_radec, header_3to2
from .beam import ConvFun
from .radec import _tight_ra
from . import conf
from .corr_vel import freq2vel, vel2freq
from .parallel_fill_grid import optimized_search_around_sky
from ..utils.io import get_nB
from ..utils.io import replace_nB

from multiprocessing import RawArray
from multiprocessing import Process, Queue

try:
    import bottleneck as bn
    NANMEAN = bn.nanmean
    NANSUM = bn.nansum
except ImportError:
    NANMEAN = np.nanmean
    NANSUM = np.nansum


import re
def groups_print_fnames(fnames):
    from itertools import groupby
    nBs = [get_nB(fname) for fname in fnames]
    replace_nB = lambda path,s:re.sub(r'[0-9][0-1]M-', f"-M{s}"[::-1], path[::-1], count=1)[::-1]
    pairs = [(replace_nB(fname, '{Beam}'), get_nB(fname)) for fname in fnames]
    pairs.sort()
    for key, group in groupby(pairs, lambda x:x[0]):
        print(key)
        print('Beams:', ', '.join([str(pp[1]) for pp in group]))


class SpecFile():
    """
    load spec info from hdf5 file
    """
    def __init__(self, fpath, type3='freq'):

        self.fpath = fpath
        self.type3 = type3
        self.nB = get_nB(fpath)
        self.load_misc()

    def __repr__(self):
        _repr = f"File: {self.fpath}"
        return _repr

    def get_frame(self, f):
        """
        try to read rest frame from header
        """
        try:
            self.frame = f['Header'].attrs['frame']
        except:
            self.frame = ''

    def load_misc(self, ):
        with h5py.File(self.fpath, 'r') as f:
            S = f['S']
            self.ra = S['ra'][()]
            self.dec = S['dec'][()]
            self.mjd = S['mjd'][()]
            self.freq = S['freq'][()]
            self.get_frame(f)
        # not here
        # self.ra = _tight_ra(self.ra)

        self.nchan_ori = len(self.freq)
        self.nspec_ori = len(self.ra)

        if self.type3.upper() == 'FREQ':
            self.arr3 = self.freq
        elif self.type3.upper() == 'VOPT':
            self.arr3 = freq2vel(self.freq, vtype='optical')
        elif self.type3.upper() == 'VRAD':
            self.arr3 = freq2vel(self.freq, vtype='radio')

    def select_spec(self, ra_range, dec_range, r_add):
        """
        select spectra in ra and dec range; generate self.is_use_t; also change ra, dec
        """
        if ra_range is not None:
            ra_tmp = self.ra
            is_use_t = (ra_tmp >= (ra_range[0] - 2*r_add)) & (ra_tmp <= (ra_range[1] + 2*r_add))
            is_use_t |= ((ra_tmp >= (ra_range[0]+360 - 2*r_add)) & (ra_tmp <= (ra_range[1]+360 + 2*r_add)))
            is_use_t |= ((ra_tmp >= (ra_range[0]-360 - 2*r_add)) & (ra_tmp <= (ra_range[1]-360 + 2*r_add)))
        else:
            is_use_t = np.full(len(self.ra), True)
        if dec_range is not None:
            dec_ = self.dec
            is_use_t &= (dec_ >= (dec_range[0] - 2*r_add)) & (dec_ <= (dec_range[1] + 2*r_add))
        if np.all(is_use_t):
            self.is_use_t = None
        else:
            self.is_use_t = is_use_t
            self.ra = self.ra[is_use_t]
            self.dec = self.dec[is_use_t]


def align_chan(arr3s, inds, range3=None):
    """
    vlrs, Ta: list; arr3s[i] in descending order
    """
    ranges = np.array([np.min([i[0] for i in arr3s]), np.max([i[-1] for i in arr3s])])
    delta = (arr3s[0][0] - arr3s[0][1])/2.
    vmax, vmin = ranges.max() + delta, ranges.min()-delta
    if range3 is not None:
        vmax = min(vmax, range3[1])
        vmin = max(vmin, range3[0])
    _arr3s = []
    _inds = []

    for i, j in zip(arr3s, inds):
        is_use = (i >= vmin) & (i <= vmax)
        _arr3s += [i[:][is_use],]
        _inds += [j[:][is_use],]

    arr3s_len = np.array([len(i) for i in _arr3s])
    len_use = arr3s_len.min()

    arr3s = [i[:len_use] for i in _arr3s]
    inds = [i[:len_use] for i in _inds]

    return arr3s, inds


class Imaging():
    """
    """
    def __init__(self, args, inplace_args=False):

        args = args if inplace_args else copy.deepcopy(args)
        if len(args.bwidth) == 1:
            args.bwidth = args.bwidth*2
        self.args = args

        # Set temporary directory for multiprocessing
        self._setup_temp_dir()

        self.check_outname()

    def _setup_temp_dir(self):
        """
        Configure temporary directory for multiprocessing.
        Only enabled when --share_mem is True.
        """
        args = self.args

        # Only set temp_dir when share_mem is enabled
        if not getattr(args, 'share_mem', False):
            return

        # Only apply on Linux
        if sys.platform != 'linux':
            print("Warning: --share_mem is only supported on Linux, ignoring temp_dir setting")
            return

        temp_dir = getattr(args, 'temp_dir', '/dev/shm')

        # Check if temp_dir exists and is writable
        if not os.path.exists(temp_dir):
            print(f"Error: Temporary directory '{temp_dir}' does not exist.")
            print("Please create the directory or specify a different one with --temp_dir")
            sys.exit(1)

        if not os.access(temp_dir, os.W_OK):
            print(f"Error: Temporary directory '{temp_dir}' is not writable.")
            print("Please check permissions or specify a different directory with --temp_dir")
            sys.exit(1)

        import tempfile
        tempfile.tempdir = temp_dir
        print(f"Using temporary directory for multiprocessing: {temp_dir}")
        if temp_dir == '/dev/shm':
            print("  (in-memory, fast but limited by RAM)")
        else:
            print("  (disk-based, slower but more space)")

    def check_outname(self):
        args = self.args
        if os.path.exists(args.outname):
            if args.force:
                print(f"will overwrite the existing output file {args.outname}")
            else:
                print(f"File exists {args.outname}")
                print("exit... Use ' -f ' to overwrite it or change outname.")
                sys.exit(0)


    def gen_fpaths(self, ):
        """fpaths from fpattern"""
        args = self.args

        fpaths = []
        for fpattern in args.fpatterns:
            fpaths += glob(fpattern)
        fpaths = list(set(fpaths))
        fpaths.sort()
        if len(fpaths) == 0:
            raise(ValueError('no file input'))
        try:
            groups_print_fnames(fpaths)
        except:
            print(*fpaths, sep='\n')
        self.fpaths = fpaths

    def gen_specfiles(self, ):
        """
        read basic info from fpath, and align chan axis
        """
        args = self.args

        specfiles = [SpecFile(fpath, args.type3) for fpath in self.fpaths]
        [s.select_spec(args.ra_range, args.dec_range, (args.r_cut+max(args.bwidth))/3600)
         for s in specfiles]
        specfiles = np.asarray(specfiles, dtype='object')

        ## align chan axis
        is_desc = specfiles[0].arr3[1] - specfiles[0].arr3[0] < 0
        if is_desc:
            arr3s = [sf.arr3 for sf in specfiles]
            inds = [np.arange(sf.nchan_ori) for sf in specfiles]
        else:
            arr3s = [sf.arr3[::-1] for sf in specfiles]
            inds = [np.arange(sf.nchan_ori)[::-1] for sf in specfiles]
        arr3s, inds = align_chan(arr3s, inds, args.range3)
        for i, (ind_use, arr3) in enumerate(zip(inds, arr3s)):
            if not is_desc:
                ind_use = ind_use[::-1]
                arr3 = arr3[::-1]
            if len(ind_use) != specfiles[i].nchan_ori:
                specfiles[i].ind_use_chan = ind_use
            else:
                specfiles[i].ind_use_chan = None
            specfiles[i].arr3 = arr3
        ## align end
        self.specfiles = specfiles

    def check_spec_key(self,):
        args = self.args
        with h5py.File(self.specfiles[0].fpath, 'r') as f:
            if args.key is None:
                if 'flux' in f['S'].keys():
                    args.key = 'flux'
                elif 'Ta' in f['S'].keys():
                    args.key = 'Ta'
                elif 'Power' in f['S'].keys():
                    args.key = 'Power'
                else:
                    raise(ValueError('no flux or temperature information'))
            print(f"Use field `{args.key}`")

    def check_axis3(self,):
        """"""
        args = self.args
        #check channel align
        arr3_stack = np.vstack([sf.arr3 for sf in self.specfiles])
        std_ = np.std(arr3_stack, axis=0, dtype=np.float64) # single precision can be inaccurate
        print(f'{args.type3} dispersion (std) at the same channel is between {np.nanmin(std_)} and {np.nanmax(std_)}.')
        if np.nanmax(std_)>0.1:
            raise(ValueError('vel dispersion (std) at the same channel is two large'))

    def get_history(self,):
        args = self.args
        #record history
        histories = []
        try:
            from .._version import get_versions
            histories = ['version: ' + get_versions()['version']]
            histories += [f"{k}: {v}" if k !='fpatterns' else "args" for k,v in args.__dict__.items()]
        except:
            pass
        histories += self.fpaths
        return histories

    def change_header(self,):
        """
        change header if args.wcs_from
        """
        args = self.args
        if args.wcs_from is not None:
            print(f'use the wcs sky coordinates parameters from {args.wcs_from}')
            from astropy.io import fits
            fa = fits.open(args.wcs_from)
            header2 = fa[0].header
            for key in self.header.keys():
                if key[-1:] in ['1', '2'] or key == 'LONPOLE' or key == 'LATPOLE':
                    print(f'replacing {key}')
                    try:
                        self.header[key] = header2[key]
                    except:
                        print(f'replace {key} fail')

    def gen_header_grid(self, ):
        """

        """
        args = self.args

        self.ra_stack = ra_stack = _tight_ra(np.hstack([sf.ra for sf in self.specfiles]))
        self.dec_stack = dec_stack = np.hstack([sf.dec for sf in self.specfiles])
        if args.ra_range is None:
            args.ra_range = ra_stack.min(), ra_stack.max()
        if args.dec_range is None:
            args.dec_range = dec_stack.min(), dec_stack.max()

        frame = self.specfiles[0].frame

        #
        print('generating WCS header')
        print('ra range:', args.ra_range,' deg')
        print('dec range:', args.dec_range,' deg')
        print('bwidth:', *args.bwidth, ' arcsec')
        self.header= gen_header(args.ra_range, args.dec_range,
                                args.bwidth[0]/3600, args.bwidth[1]/3600,
                                self.specfiles[0].arr3, args.proj, args.type3, frame,
                                histories=self.get_history(),
                                beam_fwhw = args.beam_fwhw,
                                ra_center=args.wcs_ra_center, dec_center=args.wcs_dec_center)
        if args.key == 'flux':
            self.header["BUNIT"] = 'Jy/beam'
        elif args.key == 'Ta':
            self.header["BUNIT"] = 'K'
        # args.wcs_from
        self.change_header()
        # gen grid points
        self.ra_grid, self.dec_grid = gen_grid_radec(self.header)



    def fill_grid(self, ):
        args = self.args

        cata = SkyCoord(self.ra_stack, self.dec_stack, unit=(u.degree, u.degree))
        grid = SkyCoord(self.ra_grid, self.dec_grid, unit=(u.degree, u.degree))

        use_optimized = getattr(args, 'use_optimized_search', True)

        if use_optimized:
            print(f'searching spectra in r_cut for each grid (optimized, n_workers={args.nproc})...')
            # optimized_search_around_sky(coords1, coords2) returns (idx1, idx2, d2d, d3d)
            # cata.search_around_sky(grid) calls search_around_sky(grid, cata) internally
            # So we need to call optimized_search_around_sky(grid, cata) to match
            # compute_3d=False to skip unused 3D distance calculation
            self.ind_g, self.ind_cata, self.d2d, _ = optimized_search_around_sky(
                grid.ravel(), cata.ravel(), args.r_cut*u.arcsec,
                n_workers=args.nproc, compute_3d=False, verbose=False
            )
        else:
            print('searching spectra in r_cut for each grid (original astropy)...')
            # cata.search_around_sky(grid) returns (idx_grid, idx_cata, d2d, d3d)
            self.ind_g, self.ind_cata, self.d2d, _ = cata.ravel().search_around_sky(
                grid.ravel(), args.r_cut*u.arcsec
            )

    def gen_scale_beams(self,):
        args = self.args
        import json
        if args.scale_beams_file is not None:
            beams_c = json.load(open(args.scale_beams_file))
#             # two dim Volume = 2*pi*A*sigma^2
#             self.scale_beams = (beams_c['beam_fwhw_synt'] / np.array(beams_c['beam_fwhw_19']))**2
            self.scale_beams = np.array(beams_c['scale_beams_19'])
            assert len(self.scale_beams) == 19
            args.beam_fwhw = beams_c['beam_fwhw_synt'] # arcmin
        else:
            self.scale_beams = None

    def set_r_cut(self,):
        args = self.args
        if args.r_cut is None:
            cf = ConvFun(args.method, args.beam_fwhw,
                         gaussian_fwhw=args.gaussian_fwhw,
                         bsize=getattr(args, 'bsize', None), gsize=getattr(args, 'gsize', None))
            args.r_cut = getattr(cf, 'dis_cut_suggest')*60 # to arcsec
        print(f'r_cut: {args.r_cut} arcsec')


    def gen_conv_weis(self,):
        args = self.args
        cf = ConvFun(args.method, args.beam_fwhw,
                     gaussian_fwhw=args.gaussian_fwhw,
                     bsize=getattr(args, 'bsize', None), gsize=getattr(args, 'gsize', None))
        self.conv_weis = cf(self.d2d.arcmin)
        self.conv_obj = cf


    def save_cubes(self):
        """
        Save data cubes to FITS files sequentially to minimize memory usage.

        This method creates cube views on-demand and immediately deletes the underlying
        pixel arrays after each save, avoiding memory spikes from keeping both in memory.

        Memory behavior notes:
        - Data is computed in float64 for accuracy during accumulation/normalization
        - Saved as float32 to reduce file size (standard for FITS cubes)
        - astype('float32') creates a temporary copy in RAM, regardless of whether
          the source data is in RAM or disk-based shared memory (RawArray with temp_dir)
        - Sequential save-and-delete approach minimizes peak RAM usage by only having
          one float32 conversion in memory at a time
        - When share_mem=True: source data may be on disk (temp_dir), but astype()
          still allocates the float32 copy in RAM. This is unavoidable but brief.
        """
        args = self.args
        from astropy.io import fits
        import gc

        # Calculate shape for reshaping pixel arrays into cubes
        _shape = self.ra_grid.shape + self.pixel_data.shape[-1:]

        # Apply beam correction to pixel_data before reshaping
        # self.data_cube is a view, so modifying it modifies self.pixel_data
        self.data_cube = self.pixel_data.reshape(_shape).transpose(2,0,1)
        # Apply beam correction (modifies self.header and self.data_cube)
        # Must be done before creating _header (header copy) and saving other cubes
        # Note: Only pixel_data needs correction; pixel_finite_nums_chan and pixel_weis_sum_chan do not.
        self.apply_beam_correction()

        _header = copy.deepcopy(self.header)
        _header['BUNIT'] = ''

        # Save nums_chan_cube first and delete pixel_finite_nums_chan immediately
        outname_ = '.'.join(args.outname.split('.')[:-1]) + '-count.fits'
        nums_chan_cube = self.pixel_finite_nums_chan.reshape(_shape).transpose(2,0,1)
        hdu = fits.PrimaryHDU(nums_chan_cube, header=_header)
        print(f'Saving to {outname_}.')
        hdu.writeto(outname_, overwrite=True)
        print(f'Saved count cube.')
        del nums_chan_cube
        del self.pixel_finite_nums_chan  # Free underlying memory
        gc.collect()  # Force garbage collection to ensure memory is freed


        # Save main data cube and delete pixel_data immediately
        hdu = fits.PrimaryHDU(self.data_cube.astype('float32'), header=self.header)
        print(f'Saving to {args.outname}.')
        hdu.writeto(args.outname, overwrite=args.force)
        print(f'Saved data cube.')
        del self.data_cube
        del self.pixel_data  # Free underlying memory
        gc.collect()  # Force garbage collection to ensure memory is freed

        # Save weights cube and delete pixel_weis_sum_chan immediately
        outname_ = '.'.join(args.outname.split('.')[:-1]) + '-weights.fits'
        weis_chan_cube = self.pixel_weis_sum_chan.reshape(_shape).transpose(2,0,1)
        hdu = fits.PrimaryHDU(weis_chan_cube.astype('float32'), header=_header)
        print(f'Saving to {outname_}.')
        hdu.writeto(outname_, overwrite=True)
        print(f'Saved weights cube.')
        del weis_chan_cube
        del self.pixel_weis_sum_chan  # Free underlying memory
        del self.pixel_specs_nums  # Clean up remaining arrays
        gc.collect()  # Force garbage collection to ensure memory is freed


    def init_out(self, DataType='float64'):
        args = self.args
        npixel = np.prod(self.ra_grid.shape)
        # init
        shape = (npixel, len(self.specfiles[0].arr3))
        if args.share_mem:
            # ctype: https://docs.python.org/3/library/array.html#module-array
            if DataType == 'float64':
                shm_dtype = 'd'
            elif DataType == 'float32':
                shm_dtype = 'f'
            else:
                raise(ValueError('DataType not support'))
            self.DataType = DataType

            shm = RawArray(shm_dtype, int(np.prod(shape)))
            self.pixel_data = np.frombuffer(shm, dtype=DataType).reshape(shape)
            self.pixel_data[:] = 0
            #
            shm = RawArray(shm_dtype, int(np.prod(shape)))
            self.pixel_weis_sum_chan = np.frombuffer(shm, dtype=DataType).reshape(shape)
            self.pixel_weis_sum_chan[:] = 0
            # deal with nan value
            shm = RawArray('i', int(np.prod(shape)))
            self.pixel_finite_nums_chan = np.frombuffer(shm, dtype='int32').reshape(shape)
            self.pixel_finite_nums_chan[:] = 0

        else:
            self.pixel_data = np.zeros(shape, dtype=DataType)
            #
            self.pixel_weis_sum_chan = np.zeros(shape, dtype=DataType)
            #
            self.pixel_finite_nums_chan = np.zeros(shape, dtype='int32')
        self.pixel_specs_nums = np.zeros(npixel, dtype='int32')

    def apply_beam_correction(self, ):
        args = self.args
        new_beam_fwhw = self.conv_obj.new_beam_fwhw
        if args.apply_beam_correction:
            print(f"Modifying BMAJ and BMIN in header from {self.header['BMAJ']*60, self.header['BMIN']*60} arcmin to {new_beam_fwhw, new_beam_fwhw} arcmin, ",
                 "to reflect the spatial resolution change due to the convolution.")
            self.header['BMAJ'] = new_beam_fwhw/60 # deg
            self.header['BMIN'] = new_beam_fwhw/60 # deg

            # multiple the ratio factor only when BUNIT is Jy/beam
            print("If the BUNIT is Jy/beam, a ratio factor is needed to correct the data.")
            if self.header["BUNIT"].lower() == 'jy/beam':
                print(f"The BUNIT is Jy/beam, multiplying the beam ratio factor {self.conv_obj.beam_factor}")
                self.data_cube *= self.conv_obj.beam_factor
            else:
                print("The BUNIT is not Jy/beam, no need to multiply the beam ratio factor")
                pass
        else:
            self.header['BMAJ_new'] = new_beam_fwhw/60
            self.header['BMIN_new'] = new_beam_fwhw/60


    def __call__(self, ):
        """
        """
        # prepare
        args = self.args
        step = args.step

        self.gen_fpaths()
        self.gen_scale_beams() # also change args.beam_fwhw
        self.set_r_cut()
        self.gen_specfiles()
        self.check_spec_key()
        self.check_axis3()
        self.gen_header_grid()
        self.fill_grid()
        self.gen_conv_weis()

        ## todo: consider other weights
        weis = self.conv_weis

        self.init_out()
        # loop step
        nspecs = np.array([len(sf.ra) for sf in self.specfiles])
        starts = np.arange(0, len(self.specfiles), step)
        for i, start in enumerate(starts):
            print(f"Part [{i+1}/{len(starts)}]")
            sfs = self.specfiles[start:start+step]
            # index in current specs
            ind_cata_the = self.ind_cata - np.sum(nspecs[:start])
            # select current ind_g, ind_cata, wei,
            is_ = (ind_cata_the < np.sum(nspecs[start:start+step])) & (ind_cata_the >=0)
            ind_cata_the_use = ind_cata_the[is_]
            # if empty
            if len(ind_cata_the_use) == 0:
                # notice: make sure the skip not affect the following steps
                continue
            ind_g_use = self.ind_g[is_]
            weis_use = weis[is_]
            # check
            assert (self.dec_stack[self.ind_cata[is_]] == np.hstack([sf.dec for sf in sfs])[ind_cata_the_use]).all()

            # Note: CalcSpecPixel's share_mem is temporarily disabled (set to False)
            # to avoid potential issues with shared memory in multiprocessing.
            # The Imaging class's share_mem (in init_out) is still functional.
            csp = CalcSpecPixel(sfs, key=args.key, polar=args.polar, scale_beams=self.scale_beams, share_mem=False)
            res = csp(ind_g_use, ind_cata_the_use, weis_use, n_worker=args.nproc)
            ii = res[0]
            self.pixel_data[ii] += res[1]
            self.pixel_weis_sum_chan[ii] += res[2]
            self.pixel_specs_nums[ii] += res[3]
            self.pixel_finite_nums_chan[ii] += res[4]
        # normalization
        old_set = np.seterr()
        np.seterr(invalid='ignore')
        self.pixel_data /= self.pixel_weis_sum_chan
        self.pixel_data[self.pixel_weis_sum_chan==0] = np.nan
        if args.frac_finite_min > 0:
            self.pixel_data[self.pixel_finite_nums_chan < (self.pixel_specs_nums*args.frac_finite_min)[:, None]] = np.nan
        np.seterr(**old_set)

        # Save each cube directly and delete underlying pixel arrays immediately
        # to minimize memory usage (especially important with share_mem)
        self.save_cubes()


def _get_start_stop(arr,arr_in):
    """
    arr_in: array
       sorted
    """
    if not np.all(arr_in[:-1] <= arr_in[1:]):
        raise(ValueError('arr_in should be sorted'))
    inds_left= np.searchsorted(arr_in, arr, side='left',)
    inds_right=np.searchsorted(arr_in, arr, side='right',)
    return inds_left, inds_right


class CalcSpecPixel():
    def __init__(self, sfs, *, key='flux', polar='M', scale_beams=None, share_mem=False):
        """
        key:
        polar: 'M','XX','YY'
        """
        # init shared specs
        self.polar = polar
        self.scale_beams = scale_beams
        self.share_mem = share_mem
        self.ns = ns = [len(sf.ra) for sf in sfs]

        DataType = 'float32' # use same with that in file, i.e. float32, will convert to float64 in self.calc
        specs_shape = (np.sum(ns), len(sfs[0].arr3))
        if self.share_mem:
            # ctype: https://docs.python.org/3/library/array.html#module-array
            if DataType == 'float64':
                shm_dtype = 'd'
            elif DataType == 'float32':
                shm_dtype = 'f'
            else:
                raise(ValueError('DataType not support'))
            self.shm_specs = RawArray(shm_dtype, int(np.prod(specs_shape)))
            self.specs = np.frombuffer(self.shm_specs, dtype=DataType).reshape(specs_shape)
        else:
            self.shm_specs = self.specs = np.zeros(specs_shape, dtype=DataType)

        self.DataType = DataType
        self.key = key
        self.load_data(sfs)

    def load_data(self, sfs):

        for sf, s in zip(sfs, np.cumsum([0, *self.ns[:-1]])):
            f = h5py.File(sf.fpath)
            print(f"loading: {sf.fpath}", end='\033[K\r')
            f_spec = f['S'][self.key]
            if self.polar == 'XX':
                sli = 0
            elif self.polar == 'YY':
                if f_spec.shape[0] < 2:
                    raise(ValueError(f'polar method is YY, but spectra have only one polar.'))
                sli = 1
            else:
                sli = slice(None)
            if sf.is_use_t is not None:
                ind_use_t = np.where(sf.is_use_t)[0]
                if sf.ind_use_chan is not None:
                    if (np.diff(sf.ind_use_chan) == 1).all():
                        _specs = f_spec[sli, ind_use_t, sf.ind_use_chan[0]:sf.ind_use_chan[-1]+1]
                    else:
                        raise(ValueError('select chans fail'))
                else:
                    _specs = f_spec[sli, ind_use_t]
            else:
                if sf.ind_use_chan is not None:
                    if (np.diff(sf.ind_use_chan) == 1).all():
                        _specs = f_spec[sli, :, sf.ind_use_chan[0]:sf.ind_use_chan[-1]+1]
                    else:
                        raise(ValueError('select chans fail'))
                else:
                    _specs = f_spec[sli]
            if self.polar == 'M':
                _specs = np.mean(_specs, axis=0) # only two float32 to average
            elif self.polar in ['XX', 'YY']:
                ...
            elif self.polar == 'B':
                ...
            else:
                raise(ValueError(f'polar method not support {self.polar}'))
            if self.scale_beams is not None:
                _specs *= self.scale_beams[sf.nB-1]
                print(self.scale_beams[sf.nB-1])
            self.specs[s:s+len(_specs)] = _specs.astype(self.DataType)
            f.close()
        print('')

    @staticmethod
    def calc(q, s, e,
             inds, inds_dtype, inds_shape,
             specs, specs_dtype, specs_shape,
             weis, weis_dtype, weis_shape):

        if not isinstance(inds, np.ndarray):
            inds = np.frombuffer(inds, dtype=inds_dtype).reshape(inds_shape)
        if not isinstance(specs, np.ndarray):
            specs = np.frombuffer(specs, dtype=specs_dtype).reshape(specs_shape)
        if not isinstance(weis, np.ndarray):
            weis = np.frombuffer(weis, dtype=weis_dtype).reshape(weis_shape)

        _SUM = NANSUM
        val_c = []
        num = []
        num_finite_chan = []
        wei_sum_chan = []
        for s_, e_ in zip(s, e):
            val = specs[inds[s_:e_]]
            wei = weis[s_:e_] # wei is align with inds
            wei = wei.reshape((-1,)+(1,)*(val.ndim-1))
            # deal with nan value
            is_finite = np.isfinite(val)
            val_c += [_SUM(val.astype('float64')*wei, axis=0),]
            num += [len(wei),]
            num_finite_chan += [np.sum(is_finite, axis=0),]
            wei_sum_chan += [_SUM(wei*is_finite, axis=0),]
        res = np.vstack(val_c), np.vstack(wei_sum_chan), np.hstack(num), np.vstack(num_finite_chan)

#         import time
#         print('ssssss')
#         time.sleep(15)
#         res = [None, None, None]
        if q is None:
            # not multiprocessing
            return res
        else:
            q.put(res)

    def __call__(self, ind_g_use, ind_cata_the_use, weis_use, n_worker=3):

        if self.share_mem:
            # ctype: https://docs.python.org/3/library/array.html#module-array
            shm1 = RawArray('q', int(np.prod(ind_cata_the_use.shape)))
            shm1_arr = np.frombuffer(shm1, dtype='int64').reshape(ind_cata_the_use.shape)
            shm1_arr[:] = ind_cata_the_use
            shm2 = RawArray('d', int(np.prod(weis_use.shape)))
            shm2_arr = np.frombuffer(shm2, dtype='float64').reshape(weis_use.shape)
            shm2_arr[:] = weis_use
        else:
            shm1 = shm1_arr = ind_cata_the_use
            shm2 = shm2_arr = weis_use

        print('spread these spectra into grids')
        ind_g_use_uni = np.unique(ind_g_use)
        start, stop = _get_start_stop(ind_g_use_uni, ind_g_use)

#         print('dist')
#         import time
#         time.sleep(10)
        n_worker = min(len(start), n_worker)
        if n_worker > 1:
            print(f"assign gird to {n_worker} processes")
            ps = []
            qs = []
            start_list = np.array_split(start, n_worker)
            stop_list = np.array_split(stop, n_worker)
            for start_, stop_ in zip(start_list, stop_list):
                q = Queue()
                p = Process(target=self.calc, args=(q, start_, stop_,
                                                   shm1, shm1_arr.dtype, shm1_arr.shape,
                                                   self.shm_specs, self.specs.dtype, self.specs.shape,
                                                   shm2, shm2_arr.dtype, shm2_arr.shape),)
                ps += [p]
                qs += [q]
            for p in ps:
                p.start()
            res = [q.get() for q in qs]
            for p in ps:
                p.join()  # need after q.get()
            print('gathering results')
            pixel_data = np.vstack([r[0] for r in res])
            pixel_weis_sum_chan = np.vstack([r[1] for r in res])
            pixel_specs_nums = np.hstack([r[2] for r in res])
            pixel_finite_nums_chan = np.vstack([r[3] for r in res])
        else:
            pixel_data, pixel_weis_sum_chan, pixel_specs_nums, pixel_finite_nums_chan = self.calc(None, start, stop,
                                                   shm1, shm1_arr.dtype, shm1_arr.shape,
                                                   self.shm_specs, self.specs.dtype, self.specs.shape,
                                                   shm2, shm2_arr.dtype, shm2_arr.shape)

        return ind_g_use_uni, pixel_data, pixel_weis_sum_chan, pixel_specs_nums, pixel_finite_nums_chan


from argparse import Namespace

## exampler parameters
args = Namespace()

args.ra_range = [20.254, 27.25]
args.dec_range = None

args.wcs_ra_center = None
args.wcs_dec_center = None

args.range3 = [-2000, 1500]
args.type3 = 'vrad'
args.bwidth = [60,]
args.proj = 'SIN'
args.r_cut = 90 # arcsec
##
args.method = 'bessel_gaussian'
args.beam_fwhw = 2.9 # arcmin; beam Full width at half maximum
args.gaussian_fwhw = None
args.bsize = None # for Bessel in Bessel*gauss or sin in sin*gauss
args.gsize = None # for gauss in Bessel*gauss or sin*gauss

args.fpatterns = ['/home/jyj/jingyj/pipeline_test/M31_v4_highz/S2/med/*-bld-fc-ds.hdf5', ]

args.outname = 'test.fits'
args.force = True

args.nproc = 5
args.step = 19
args.share_mem = False

args.key = None
args.wcs_from = '/home/jyj/jingyj/M33/M33_local_areciob.fits'

args.scale_beams_file = None

args.polar = 'M'

args.frac_finite_min = 0.01




"""
hifast.flux
===========

Flux calibration module.

This module converts the unit of spectra from Antenna Temperature (Ta, Kelvin) to Flux Density (Jy).
It supports two calibration methods:
1.  **Noise Diode Calibration**: Uses a calibrator source (e.g., 3C48) observed with a noise diode to determine the gain.
2.  **Pre-measured Gain**: Uses a standard gain curve (e.g., Jiang et al. 2020) based on zenith angle.
"""

__all__ = ['IO', 'parser']


from .utils.io import *


def create_parser():
    sep_line = '##'+'#'*70+'##'
    parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                            description='Convert spectra from Antenna Temperature (Ta) to Flux Density (Jy).')
    add_common_argument(parser)

    # --- Input/Output ---
    group = parser.add_argument_group('Input/Output')
    group.add_argument('fpath', metavar='FILE',
                        help='Input spectra file path (HDF5 format, typically with units in Ta).')

    # --- Calibration Method: Noise Diode ---
    group = parser.add_argument_group('Calibration: Noise Diode')
    group.add_argument('--cbr_store', '--cali_fname', dest='cbr_store', default='none', metavar='FILE/DIR',
                       help='Path to the calibrator file or directory containing calibrator files. If specified, noise diode calibration is used.')
    group.add_argument('--cbr_name', default='*', metavar='NAME',
                       help='Name pattern of the calibrator source (e.g., "3C48", "3C286"). Used to filter files if a directory is provided.')
    group.add_argument('--only_use_19beams', type=bool_fun, choices=[True, False], default='False',
                       help='If True, only use calibrator data that has all 19 beams.')
    group.add_argument('--fix_diff_tcal', type=bool_fun, choices=[True, False], default='True',
                       help='Correct for Tcal differences between the target observation and the calibrator observation.')
    group.add_argument('--fix_diff_ZA', type=bool_fun, choices=[True, False], default='False',
                       help='Correct for gain differences due to Zenith Angle (ZA) differences between target and calibrator.')

    # --- Calibration Method: Pre-measured Gain ---
    group = parser.add_argument_group('Calibration: Pre-measured Gain')
    group.add_argument('--pre_measured', default='Jiang2020', choices=['Jiang2020', 'Liu2024'],
                       help='Select the pre-measured gain curve to use if `--cbr_store` is not set. "Jiang2020" (arXiv:2002.01786) is the standard.')
    group.add_argument('--Atemp', '--Tamb', type=float, metavar='DEG_C',
                       help='Ambient temperature [Celsius]. Required for "Liu2024" model correction. Ignored for "Jiang2020".')

    # --- General Settings ---
    group = parser.add_argument_group('General Settings')
    group.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')], metavar=('MIN', 'MAX'),
                        help='Limit the frequency range to process [MHz]. Followed by two values (min, max).')
    group.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                        help="Skip checking or adding RA/DEC coordinates (use with caution).")
    
    # Internal/Hidden
    parser.add_argument('--flux', type=bool_fun, choices=[True], default='True', not_in_write_out_config_file=True,
                       help=argparse.SUPPRESS)

    return parser

parser = create_parser()


class IO(BaseIO):
    ver = 'old'

    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-flux'
        return fpart

    def gen_s2p_out(self,):
        args = self.args
        s2p = self.s2p[:]
        from .core.flux import FluxCali
        print('Flux calibrating ...')
        if args.fix_diff_tcal:
            tcal_spec = self.fs['Tcal'][:]
            if self.is_use_freq is not None:
                tcal_spec = tcal_spec[:, self.is_use_freq]
        else:
            tcal_spec = None
#         if args.cbr_store != 'none' and args.cbr_store is not None:
#             print(f'using {args.cbr_store} ...')

        fcali = FluxCali(self.nB, self.freq, cbr_store=args.cbr_store, cbr_name=args.cbr_name, tcal_spec=tcal_spec,  only_use_19beams=args.only_use_19beams,
                         pre_measured=args.pre_measured, ATemp=args.Atemp,
                         mjd=self.mjd, ra=self.ra, dec=self.dec, fix_diff_ZA=args.fix_diff_ZA)
        self.s2p_out = fcali(s2p)

        if args.cbr_store is not None and args.cbr_store != 'none':
            self.Header['Calibrater_fpath'] = fcali.cbr_fpath
        else:
            self.Header['pre_measured_fluxgain'] = args.pre_measured


if __name__ == '__main__':
    dests_hide = ['flux',]
    hide_paras(parser, dests_hide)

    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    args_from = del_paras_in_string(args_from, dests_hide)
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()





__all__ = ['get_baseline', 'get_baseline_mp', 'BL_base', 'BL_PLS', 'BL_arPLS', 'BL_Gauss', 'BL_Chebyshev', 'BL_poly',
           'BL_sin_poly', 'BL_sin_poly_2', 'piecewise_polyfit', 'BL_knpoly', 'BL_spline', 'BL_knspline', 'BL_asPLS',
           'BL_masPLS', 'get_exclude_fun', 'sub_baseline']


from functools import lru_cache
import sys

import scipy.sparse as sparse
from scipy.sparse import linalg
from numpy.linalg import norm
import numpy as np
import scipy.interpolate as interp
from scipy import ndimage
from scipy import optimize
from threadpoolctl import ThreadpoolController

from astropy.stats import sigma_clipped_stats

from ..utils.misc import average_every_n, smooth1d, smooth1d_fft, extend_Trues

import os
import warnings
warnings.filterwarnings("ignore", r'overflow encountered in exp')
warnings.filterwarnings("ignore", r'The fit may be poorly conditioned')

try:
    import bottleneck as bn
    MEDIAN = bn.median
    MEAN = bn.nanmean
    STD = bn.nanstd
except ImportError:
    MEDIAN = np.median
    MEAN = np.mean
    STD = np.std


def get_baseline(x, ys, axis=None, *,
                 s_method='none', s_sigma=3, average_every=None, exclude=None,
                 exclude_add='none',
                 method='arPLS', bl_para=None,
                 verbose=False, return_f=False, check=True,
                 interp_nan=False,
                 ):
    """
    subtract the baseline of spectra with XX and YY polar.
    Parameters：
    ---------------
    x: array, shape (n,); freq or vel.
    ys: array, T or Jy
    axis: axis along fitting baseline
    s_sigma, s_method: if s_sigma is not 'none', smoothing ys along axis using s_method ('gaussian' or 'boxcar')
    average_every: int; if set as n and n>1, then average ys along axis every n points
    return_f: If Fasle, return baseline, else return (basline, ys_processed, x_processed, weights)
    check: if True, check if there are nan or inf in ys, if there are, add them to exclude if intep_nan is False.
    intep_nan: if True and check is True, replace nan values with interp.interp1d but not add to exclude.
    """
    if x is not None:
        x = np.array(x, dtype='float')
    if exclude is not None:
        if exclude.shape != ys.shape:
            raise(ValueError('exclude should has same shape with ys'))
    # check if ys is one dim
    if ys.ndim == 1:
        ys = ys[:, None]
        if exclude is not None:
            exclude = exclude[:, None]
        ch = True
        axis = 0
    else:
        ch = False

    # smooth ys along axis
    if s_method is not None and s_method != 'none':
        if s_method in ['boxcar', 'gaussian', 'median']:
            ys = smooth1d(ys, s_method, s_sigma, axis=axis)
        elif s_method == 'fft':
            ys = smooth1d_fft(x, ys, s_sigma, axis=axis)
        else:
            raise(ValueError('not support the method'))
    else:
        ys = ys

    if average_every is not None and average_every > 1:
        drop = False
        x_ori = x
        x = average_every_n(x, average_every, drop=drop)
        ys = average_every_n(ys, average_every, axis=axis, drop=drop)
        if exclude is not None:
            exclude = average_every_n(
                exclude, average_every, axis=axis, drop=drop).astype('bool')

    if 'PLS' in method:
        bl_para = bl_para.copy()  # shallow copy
        if 'sym' in bl_para.keys():
            del bl_para['sym']
        if bl_para['deg'] > 3:
            warnings.warn('deg is too large in PLS methods and has been set to 3.')
            bl_para['deg'] = 3  # bl_para has been shallow copyed
    if 'spline' in method:
        #  also for 'knspline'
        bl_para = bl_para.copy() # shallow copy
        if 'sym' in bl_para.keys():
            del bl_para['sym']
        if bl_para['deg'] > 5:
            warnings.warn('deg is too large in spline methods and has been set to 5.')
            bl_para['deg'] = 5  # bl_para has been shallow copyed
        if method.startswith('spline'):
            if 'knots' in bl_para.keys():
                del bl_para['knots']
        if method.startswith('knspline'):
            if 'knots' not in bl_para.keys() or bl_para['knots'] is None:
                raise(ValueError('need knots for knspline- method'))
    if method.startswith('knpoly'):
        if 'knots' not in bl_para.keys() or bl_para['knots'] is None:
            raise(ValueError('need knots for knpoly- method'))
    # select baseline method

    if '-' in method:
        p1, p2 = method.split('-')
        if p1 == 'PLS':
            BL = BL_PLS(**bl_para, rew_type=p2)
        elif p1 == 'spline':
            BL = BL_spline(**bl_para, rew_type=p2)
        elif p1 == 'knspline':
            BL = BL_knspline(**bl_para, rew_type=p2)
        elif p1 == 'Chebyshev':
            BL = BL_Chebyshev(**bl_para, rew_type=p2)
        elif p1 == 'poly':
            BL = BL_poly(**bl_para, rew_type=p2)
        elif p1 == 'knpoly':
            BL = BL_knpoly(**bl_para, rew_type=p2)
        elif p1 == 'Gauss':
            BL = BL_Gauss(**bl_para, rew_type=p2)
        elif p1 == 'sin_poly':
            BL = BL_sin_poly(**bl_para, rew_type=p2)
        elif p1 == 'masPLS':
            BL = BL_masPLS(**bl_para, rew_type=p2)
        else:
            raise(ValueError('not supported method: {method}'))
    elif method == 'arPLS':
        BL = BL_PLS(**bl_para, rew_type='asym')
    elif method == 'srPLS':
        BL = BL_PLS(**bl_para, rew_type='sym')
    elif method == 'masPLS':
        BL = BL_masPLS(**bl_para, rew_type='asym')
    elif method == 'asPLS':
        BL = BL_asPLS(**bl_para)
    elif method == 'sin_poly':
        BL = BL_sin_poly(**bl_para, rew_type='asym')
    elif method == 'original':
        return ys
    else:
        raise(ValueError('not supported method: {method}'))


    # move axis to last
    ys = np.moveaxis(ys, axis, -1)
    if exclude is not None:
        exclude = np.moveaxis(exclude, axis, -1)
    shape_bak = ys.shape
    # ys = ys.reshape((-1,shape_bak[-1])) #will copy the data
    # or np.apply_along_axis
    bls = np.zeros_like(ys)
    if return_f:
        weis = np.zeros_like(ys, dtype=np.float64)
    if verbose:
        from tqdm import tqdm
        iter_ = tqdm(np.ndindex(
            *(shape_bak[:-1])), total=ys.size/shape_bak[-1], desc='CPU 0: ', mininterval=2)
    else:
        iter_ = np.ndindex(*(shape_bak[:-1]))

    for ii in iter_:
        y = ys[ii + np.s_[:, ]]
        if exclude is not None:
            _exclude = exclude[ii + np.s_[:, ]]
        else:
            _exclude = None
        if check:
            is_finite = np.isfinite(y)
            if not is_finite.all():
                y_finite = y[is_finite]
                # if all nan
                if 0 in y_finite.shape:
                    bls[ii + np.s_[:, ]] = np.nan
                    continue
                if not interp_nan:
                    # add nan to exclude and replace nan with max or any value
                    y = np.copy(y)
                    y[~is_finite] = np.max(y_finite)
                    if exclude is not None:
                        _exclude = _exclude | (~is_finite)
                    else:
                        _exclude = ~is_finite
                else:
                    # not add to exclude but repalce nan values with interp.interp1d
                    y = np.copy(y)
                    y[~is_finite] = interp.interp1d(
                        x[is_finite], y_finite, kind='linear', fill_value='extrapolate')(x[~is_finite])

            if _exclude is not None and _exclude.all():
                warnings.warn('All points in this spectrum are excluded, it\'s baseline will set as all zeros.')
                bls[ii + np.s_[:, ]] = 0
                continue
        try:
            bl = BL.fit(x=x, y=y, exclude=_exclude, exclude_add=exclude_add)
        except Exception as err:
            print('fit baseline fail, return zero array. Try other method, or there are bugs and ...')
            print('Errors output:\n')
            print(err)
            print('')
            sys.stdout.flush()
            continue # bls initialized as zeros
        bls[ii + np.s_[:, ]] = bl
        if return_f:
            weis[ii + np.s_[:, ]] = BL.wei
    # move back
    bls = np.moveaxis(bls, -1, axis)
    if return_f:
        weis = np.moveaxis(weis, -1, axis)
    if average_every is not None and average_every > 1:
        bls = interp.interp1d(x, bls, axis=axis, kind='linear',
                              fill_value='extrapolate')(x_ori)
    if return_f:
        # move back
        ys = np.moveaxis(ys, -1, axis)
        if ch:
            return bls[:, 0], ys[:, 0], x, weis[:, 0]
        else:
            return bls, ys, x, weis
    else:
        if ch:
            return bls[:, 0]
        else:
            return bls


controller = ThreadpoolController()



class get_baseline_mp(object):
    def __init__(self, n):
        # n process
        self.n = n

    @staticmethod
    def get_baseline_q(q, *args, **kwargs):
        res = get_baseline(*args, **kwargs)
        q.put(res)

    @controller.wrap(limits=1, user_api='blas')
    def __call__(self, x, ys, *, exclude=None,  **kwargs):
        """
        testing...
        only support axis=1,
        not support return_f=True
        """
        from functools import partial
        from multiprocessing import Process, Queue

        if 'axis' in kwargs.keys():
            if kwargs['axis'] != 1:
                raise(ValueError('get_baseline_mp only support axis=1 for now'))
        else:
            kwargs['axis'] = 1
        n = min(self.n, len(ys))

        ps = []
        qs = []
        if exclude is not None:
            exclude_list = np.array_split(exclude, n)
        else:
            exclude_list = [None, ] * n
        ys_list = np.array_split(ys, n)
        for ys, exclude in zip(ys_list, exclude_list):
            q = Queue()
            p = Process(target=self.get_baseline_q, args=(q, x, ys),
                        kwargs={'exclude': exclude, **kwargs})
            if 'verbose' in kwargs.keys():
                kwargs['verbose'] = False

            ps += [p]
            qs += [q]
        for p in ps:
            p.start()
        res = np.vstack([q.get()for q in qs])
        for p in ps:
            p.join()  # need after q.get()
        return np.stack(res)


_rew_fun_dict = {}

#@classmethod
def _reweight_asym1(self, d):
    """
    same to ``Baek et al. 2015, Analyst 140: 250-257``
    """
    if self.exclude is None:
        dn = d[d < 0]
    else:
        dn = d[(d < 0) & (~self.exclude)]
    m = MEAN(dn)
    s = STD(dn)
    wt = 1./(1 + np.exp(2 * (d-(self.offset*s-m))/s))
    return wt

_rew_fun_dict['asym1'] = _reweight_asym1

def _reweight_asym2(self, d):
    """
    similar to ``Baek et al. 2015, Analyst 140: 250-257``
    """
    if self.exclude is None:
        dn = d[d < 0]
    else:
        dn = d[(d < 0) & (~self.exclude)]
    m = MEAN(dn)
    s = STD(dn)
    wt = 1./(1 + np.exp(2 * (abs(d)-(self.offset*s-m))/s))
    return wt

_rew_fun_dict['asym2'] = _reweight_asym2

def _reweight_asym3(self, d):
    """
    similar to ``Baek et al. 2015, Analyst 140: 250-257``
    """
    if self.exclude is None:
        dn = d[d < 0]
    else:
        dn = d[(d < 0) & (~self.exclude)]
    m = MEAN(dn)
    s = STD(dn)
    wt = np.zeros_like(d)
    wt[abs(d) < self.offset*s-m] = 1
    return wt

_rew_fun_dict['asym3'] = _reweight_asym3

def _reweight_sym1(self, d):
    """
    """
    dn = d if self.exclude is None else d[~self.exclude]
    mean, median, stddev = sigma_clipped_stats(dn)
    m = median
    s = stddev/2.
    wt = 1./(1 + np.exp(2 * (abs(d)-(self.offset*s-m))/s))
    return wt

_rew_fun_dict['sym1'] = _reweight_sym1

## set default
_rew_fun_dict['asym'] = _reweight_asym1
_rew_fun_dict['sym'] = _reweight_sym1


class BL_base(object):
    """
    (automatic) Baseline fit
    Parameters
    ----------
    offset:
    deg:
    rew_type:
    ratio:
    niter:
    rew: deprecated, use niter=1 instead
    """

    def __init__(self, *, offset=2, deg=3, ratio=0.01, niter=100, rew_type='asym', rew=True, sym=False):

        self.offset = offset
        self.deg = deg
        self.ratio = ratio
        self.niter = niter
        BL_base._reweight = _rew_fun_dict[rew_type]
        if not rew:
            self.niter = 1

    def _fit(self, x, y, w):
        ...
        return y

    def fit(self, *, x=None, y=None, exclude=None, exclude_add='none', wei=None):
        """
        iteration

        Parameters
        ----------
        x: None or array (N,)
        y: array (N,); nan value need be replaced as normal nubmer in advance, and put them in ``exclude``
        exclude: bool array (N,); masked points that will never be used in fitting baseline
        exclude_add: str; additional exclude at last step, 'none', 'auto', 'auto1', 'auto2'
        wei: array; initial weigth for each point

        """
        # Adaptation of the code in Baek et al 2015 and https://github.com/charlesll/rampy
        self.exclude = exclude
        N = len(y)
        wt = np.ones(N) if wei is None else np.copy(wei)
        if self.exclude is not None:
            wt[self.exclude] = 0
        for i in range(self.niter):
            w = wt
            z = self._fit(x, y, w)
            d = y - z
            wt = self._reweight(d)
            # also used to deal replaced nan value, need applied always
            if self.exclude is not None:
                wt[self.exclude] = 0
            # check exit condition and backup
            ratio_fit = norm(w-wt)/norm(w)
            if ratio_fit < self.ratio:
                break
        exclude2 = None
        if exclude_add == 'auto' or exclude_add == 'auto1':
            is_ = w<0.01
            if self.exclude is not None:
                is_[self.exclude] = False # not extend the input exclude range
            exclude2 = extend_Trues(is_, axis=-1, ext_frac=1/3)
        elif exclude_add == 'auto2':
            ds = ndimage.gaussian_filter1d(d, 3)
            is_ = ds > 3*STD(d[d<0])
            if self.exclude is not None:
                is_[self.exclude] = False # not extend the input exclude range
            exclude2 = extend_Trues(is_, axis=-1, ext_frac=1/3)
        if exclude2 is not None:
            if self.exclude is not None:
                exclude2 |= self.exclude
            w = wt # use wt
            w[exclude2] = 0
            z = self._fit(x, y, w)

        bl = z
        self.success = True if ratio_fit <= self.ratio else False
        self.wei = w
        self.i = i
        return bl


class BL_PLS(BL_base):
    """
    (automatic) Baseline correction using asymmetrically reweighted penalized least squares smoothing.
                     Baek et al. 2015, Analyst 140: 250-257
    Parameters
    ----------
    lam:
    offset:
    deg:
    ratio:
    niter:
    """

    def __init__(self, lam=10**8, **kwargs):
        """
        lam:
        offset:
        deg:
        ratio:
        niter:
        """
        self.lam = lam
        super().__init__(**kwargs)

    @lru_cache(maxsize=20)
    def get_H(self, N):
        """
        """
        D = sparse.csc_matrix(sparse.eye(N))
        for i in range(self.deg):
            D = D[:, 1:] - D[:, :-1] # D is a csc sparse matrix, don't use np.diff
        H = D.dot(D.transpose()).multiply(self.lam)
        return H

    def _fit(self, x, y, w):
        """
        """
        N = len(y)
#         H = getattr(self, 'H', self.get_H(N))
#         self.H = H
        H = self.get_H(N)
        W = sparse.spdiags(w, 0, N, N)
        Z_m = W + H
        z = linalg.spsolve(Z_m, w*y)
        return z

BL_arPLS = BL_PLS


class BL_Gauss(BL_base):
    """
    """
    def __init__(self, lam=200, **kwargs):
        """
        lam: here as gaussian sigma
        """
        self.lam = int(lam)
        super().__init__(**kwargs)

    def _fit(self, x, y, w):
        z = smooth1d(y*w, 'gaussian_fft', self.lam, axis=-1)
        z /= smooth1d(w, 'gaussian_fft', self.lam, axis=-1)
        return z



class BL_Chebyshev(BL_base):
    """
    (automatic) Baseline correction using Chebyshev fit
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """

    def pred(self, x):
        return np.polynomial.chebyshev.chebval(x, self.para)

    def _fit(self, x, y, w):
        para = np.polynomial.chebyshev.chebfit(x, y, self.deg, w=w)
        self.para = para
        return self.pred(x)


class BL_poly(BL_base):
    """
    (automatic) Baseline correction using polynomial fit
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """

    def pred(self, x):
        return np.polynomial.polynomial.polyval(x, self.para)

    def _fit(self, x, y, w):
        para = np.polynomial.polynomial.polyfit(x, y, self.deg, w=w)
        self.para = para
        return self.pred(x)


class BL_sin_poly(BL_base):
    """
    (automatic) Baseline correction using sin plus poly
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """

    def __init__(self, f, ptype='poly', opt_para={}, **kwarg):
        super().__init__(**kwarg)
        if np.isscalar(f):
            f = [f]
        self.f = f
        self.ptype = ptype
        self.set_opt_para(**opt_para)

    def _fun(self, x, *arg):
        arg = np.asarray(arg)
        coef = arg[-(self.deg + 1):]
        y = np.zeros_like(x)
        for _A, _f, _p in zip(*np.split(arg[:-(self.deg + 1)], 3)):
            y += _A*np.sin(2*np.pi*_f*x + _p)
        if self.ptype == 'poly':
            y += np.polynomial.polynomial.polyval(x, coef)
        elif 'cheb' in self.ptype.lower():
            y += np.polynomial.chebyshev.chebval(x, coef)
        return y

    def _err_func(self, para, x, y, w):
        residuals = y-self._fun(x, *para)
        return np.sum((residuals*w)**2, dtype='float64')

    def set_opt_para(self, method='Powell', **kwarg):
        """
        scipy.optimize.minimize parameter
        """
        self.opt_method = method
        self.opt_kwarg = kwarg

    def _fit(self, x, y, w):
        nsin = len(self.f)
        self.w = w
        std = np.std(y[w >= 0.99])
        p0 = [std] + [std/2]*(nsin-1)
        p0 += self.f + [0, ]*nsin
        p0 += [0]*(self.deg+1)

        #err_func = self._err_func(x, y, w, *arg)
        res = optimize.minimize(self._err_func, p0, args=(
            x, y, w), method=self.opt_method, **self.opt_kwarg)
        self.para = res.x
        self.fit_res = res
        return self._fun(x, *res.x)

    def pred(self, x):
        return self._fun(x, *self.para)


class BL_sin_poly_2(BL_sin_poly):
    """
    (automatic) Baseline correction using sin plus poly
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """

    def set_opt_para(self, **kwarg):
        """
        scipy.optimize.curve_fit parameter
        """
        self.opt_kwarg = kwarg

    def _fit(self, x, y, w):
        nsin = len(self.f)
        self.w = w
        std = np.std(y[w >= 1])
        p0 = [std] + [std/2]*(nsin-1)
        p0 += self.f + [0, ]*nsin
        p0 += [0]*(self.deg+1)

        res = optimize.curve_fit(
            self._fun, x, y, p0, sigma=1./w, **self.opt_kwarg)

        self.para = res[0]
        self.fit_res = res
        return self._fun(x, *self.para)



def piecewise_polyfit(x, y, weights, knots, deg, xnew):
    # Find indices at which to split data based on knots
    split_indices = np.append(np.searchsorted(x, knots), len(x))
    split_indices_new = np.append(np.searchsorted(xnew, knots), len(xnew))
    # Fit a polynomial to each section
    ynew = np.zeros(len(xnew), dtype=y.dtype)
    start = 0
    start_new = 0
    for split_index, split_index_new in zip(split_indices, split_indices_new):
        # Slice data for this section
        xs = x[start:split_index]
        ys = y[start:split_index]
        ws = weights[start:split_index]

        # Fit polynomial to this section
        p = np.polynomial.polynomial.polyfit(xs, ys, deg=deg, w=ws)

        # Evaluate polynomial for this section
        ynew[start_new:split_index_new] = np.polynomial.polynomial.polyval(xnew[start_new:split_index_new], p)

        # Update start indices for next section
        start = split_index
        start_new = split_index_new

    return ynew


class BL_knpoly(BL_base):
    """
    (automatic) Baseline correction using
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """
    def __init__(self, knots=None, **kwargs):
        """
        """
        self.knots = knots
        super().__init__(**kwargs)

    def _fit(self, x, y, w):
        return piecewise_polyfit(x, y, weights=w, knots=self.knots, deg=self.deg, xnew=x)


class BL_spline(BL_base):
    """
    (automatic) Baseline correction using
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """
    def __init__(self, lam=10**8, **kwargs):
        """
        """
        self.lam = lam
        super().__init__(**kwargs)

    def _fit(self, x, y, w):
        spl = interp.UnivariateSpline(x, y, w, k=self.deg, s=self.lam)
        return spl(x)


class BL_knspline(BL_base):
    """
    (automatic) Baseline correction using
    Parameters
    ----------
    offset:
    deg:
    ratio:
    niter:
    """
    def __init__(self, knots=None, **kwargs):
        """
        """
        self.knots = knots
        super().__init__(**kwargs)

    def _fit(self, x, y, w):
        spl = interp.LSQUnivariateSpline(x, y, t=self.knots, w=w, k=self.deg)
        return spl(x)



def _reweight_asPLS(self, d):
        # make d- and get w^t with m and s
        dn = d[d < 0]
        s = STD(dn, ddof=1)
        wt = 1.0/(1 + np.exp(self.offset * (d-s)/s))
        return wt

class BL_asPLS(BL_arPLS):
    """
    (automatic) Baseline correction using a modified version of adaptive smoothness parameter penalized least squares method.
        Feng Zhang et al. 2020, DOI: 10.1080/00387010.2020.1730908

    Parameters
    ----------
    lam:
    offset:
    deg:
    ratio:
    niter:
    """
    def __init__(self, *args, **kwrags):
        super().__init__(*args, **kwrags)
        BL_asPLS._reweight = _reweight_asPLS

    def _update_alpha(self, d):
        d_abs = np.abs(d)
        return d_abs/np.max(d_abs)

    def fit(self, *, x=None, y=None, exclude=None, wei=None):
        self.exclude = exclude
        N = len(y)
        D = sparse.csc_matrix(sparse.eye(N))
        #D= self._diff(D)

        def diff_fun(x): return x[:, 1:] - \
            x[:, :-1]  # x is a csc sparse matrix
        for i in range(self.deg):
            D = diff_fun(D)
        w = np.ones(N) if wei is None else np.copy(wei)
        alpha = np.ones(N)
        if self.exclude is not None:
            w[self.exclude] = 0
        H = D.dot(D.transpose()).multiply(self.lam)
        for i in range(self.niter):
            W = sparse.spdiags(w, 0, N, N)
            Z = W + H.multiply(alpha[:, None])
            z = linalg.spsolve(Z, w*y)
            d = y - z
            wt = self._reweight(d)
            if self.exclude is not None:
                wt[self.exclude] = 0
            # check exit condition and backup
            ratio_fit = norm(w-wt)/norm(w)
            if ratio_fit < self.ratio:
                break
            w = wt
            alpha = self._update_alpha(d)

        bl = z
        self.success = True if ratio_fit <= self.ratio else False
        self.wei = w
        return bl


class BL_masPLS(BL_asPLS, BL_arPLS):
    def __init__(self, *args, **kwrags):
        BL_arPLS.__init__(self, *args, **kwrags)
    def _update_alpha(self, d):
        d_abs = np.abs(d)
        return 1+d_abs/np.max(d_abs)


def get_exclude_fun(exclude_m):
    if exclude_m == 0:
        def exclude_fun(x):
            return extend_Trues(abs(x) > 1.*np.diff(np.percentile(x, [16, 84], axis=1), axis=0)[0][:, None, :], axis=1, ext_frac=1/3, ext_add=3)
    elif exclude_m == 1:
        def exclude_fun(x):
            return extend_Trues(abs(x) > 2.5*np.min(np.diff(np.percentile(x, [10, 50, 90], axis=1), axis=0), axis=0)[:, None, :], axis=1, ext_frac=1/3, ext_add=3)
    return exclude_fun


def sub_baseline(freq, yss, *, subtract=True, nproc=1, exclude_fun=None, is_excluded=None, exclude_add='none', inplace=False,
                 njoin=1, s_method_t='none', s_sigma_t=None,
                 method='arPLS', s_method_freq='none', s_sigma_freq=None, average_every_freq=None,
                 lam=1e8, deg=2, offset=2, ratio=0.01, niter=100, sin_f=[0.925, ], rew=True, opt_para=None,
                 verbose=True,
                 knots=None,
                 ):
    """
    freq: array, shape:(n,) ; Frequency.
    yss: array, shape:(m,n,2)

    njoin: int
    nproc:
    exclude_fun:
    s_method_t: smooth method along axis 0
    s_sigma_t:
    s_method_freq:
    s_sigma_freq:
    average_every_freq:
    method: str; baseline fitting method
    bl_para: baseline fitting parameter
    verbose:
    """
    # check
    if yss.ndim != 3:
        raise(ValueError('yss should has 3-dim'))
    if yss.shape[1] != len(freq):
        raise(ValueError('the 2nd dimension of yss should equal to the length of freq'))

    # baseline fit parameter
    bl_para = {
        'deg': deg,
        'offset': offset,
        'ratio': ratio,
        'niter': niter,
        'rew': rew,
    }
    if 'PLS' in method or 'Gauss' in method:
        bl_para['lam'] = lam
    if method.startswith('spline'):
        bl_para['lam'] = lam
    if method.startswith('knspline') or method.startswith('knpoly'):
        if knots is None:
            raise(ValueError('*-knspline methods need knots input'))
        bl_para['knots'] = knots
    if 'sin' in method:
        bl_para['f'] = sin_f
        if opt_para is not None:
            bl_para['opt_para'] = opt_para

    yss_ori = yss
    # smooth spectra along t
    if s_method_t is not None and s_method_t != 'none':
        if s_method_t in ['boxcar', 'gaussian', 'median']:
            yss = smooth1d(yss, s_method_t, s_sigma_t, axis=0)
        elif s_method_t == 'fft':
            yss = smooth1d_fft(np.arange(yss.shape[0]), yss, s_sigma_t, axis=0)
        else:
            raise(ValueError('not support the method'))
    # join
    if njoin is not None and njoin > 1:
        yss = average_every_n(yss, njoin, axis=0, drop=False)
    # add exclude
    if is_excluded is not None:
        exclude = is_excluded
        if njoin is not None and njoin > 1:
            # todo: using np.logical_and.reduce
            exclude = average_every_n(exclude, njoin, axis=0, drop=False).astype('bool')
    elif exclude_fun is not None:
        exclude = exclude_fun(yss)
    else:
        exclude = None

    if s_method_freq == 'PLS':
        # use arPLS, lam from s_sigma ( s_sigma_freq)
        yss = get_baseline_mp(nproc)(freq, yss, axis=1, method='arPLS', bl_para={
            'lam': para['s_sigma'], "offset": 2, 'deg': 2}, verbose=verbose)
        s_method_freq = None
    # get baseline
    bls = get_baseline_mp(nproc)(freq, yss, axis=1, exclude=exclude,
                                 exclude_add=exclude_add,
                                 s_method=s_method_freq, s_sigma=s_sigma_freq, average_every=average_every_freq,
                          method=method, bl_para=bl_para, verbose=verbose)

    if njoin is not None and njoin > 1:
        bls = bls[np.searchsorted(np.arange(0, yss_ori.shape[0], njoin), 1+np.arange(yss_ori.shape[0]))-1]
    if subtract:
        return yss_ori - bls
    else:
        return bls



# author: Xu Chen, Li Fujia, 2021.06
# code: Xu Chen

import numpy as np
from astropy import log
from copy import deepcopy
from tqdm import tqdm
# from matplotlib import pyplot as plt

from multiprocessing import Process, Queue
from threadpoolctl import ThreadpoolController
controller = ThreadpoolController()

def check_bottleneck():
    try:
        import bottleneck as bn
    except ModuleNotFoundError:
        bn = np
        log.warning("Install the package 'bottleneck' will help to speed up. Now use 'numpy' instead.")
    return bn

bn = check_bottleneck()


def _running_block(q, data, n, func, method,):
    
    if func == 'iter':
#         raise ValueError(np.sum(np.isnan(data)))
        
        tlen = data.shape[0]

        t1 = np.arange(tlen)-n
        t2 = np.arange(tlen)+n
        t1[t1 < 0] = 0
        t2[t2 < 2*n] = 2*n
        t1[t1 > tlen - 2*n - 1] = tlen - 2*n - 1
        t2[t2 > tlen - 1] = tlen - 1

        sw = np.zeros_like(data)
        for i in tqdm(range(tlen)):
            if method == 'median':
                sw[i] = bn.nanmedian(data[t1[i]:t2[i]], axis = 0)
            elif method == 'mean':
                sw[i] = bn.nanmean(data[t1[i]:t2[i]], axis = 0)
            
    elif func == 'smooth':
        from ..utils.misc import smooth1d
        if method == 'median':
            s_method_t = 'median'
        elif method == 'mean':
            s_method_t = 'boxcar' 
        print('Smooth ing ...')
        sw = smooth1d(data, axis=0, sigma=n, method=s_method_t)
    
    q.put(sw)

@controller.wrap(limits=1, user_api='blas')
def running_median_mp(data, n, func, method, nproc):

    n_proc = min(nproc, data.shape[1]) # freq

    ps = []
    qs = []
    i_list = np.arange(data.shape[1])
    d_list = np.array_split(i_list, n_proc, axis = 0)
    for d in d_list:
        q = Queue()
        p = Process(target = _running_block, 
                    args=(q, data[:, d, :], n, func, method,))

        ps += [p]
        qs += [q]
    for p in ps:
        p.start()
    res = np.hstack([q.get()for q in qs])

    for p in ps:
        p.join()  # need after q.get()

    return res


def running_median(data, nspec, func='iter', method = 'median', nproc = 1):
    """
    data: (mjd, freq, polar)
    nspec: moving window length
    func: 'iter' using numpy or bottleneck
          'smooth' boxcar filter
    """
    if method not in ['median', 'mean']:
        raise ValueError(f"{method} should be 'median' or 'mean'!")
    
    n = nspec // 2
    
    if func == 'iter':
        tlen = data.shape[0]
        if tlen < 2*n:
            print(f"use all specs to {method}")
        else:
            print(f"use {2*n} specs to {method}")
    elif func == 'smooth':
        print(f"use {2*n +1} specs to {method}")
        
    sw = running_median_mp(data, n, func, method, nproc)

    return sw


### minmed

def get_trange(N, npart = None, dpart = None,):
    if dpart is None:
        dpart = N // npart
    p1 = np.arange(0, N, dpart)
    p2 = np.arange(dpart, N + dpart, dpart)
    p2[p2 > N] = N
    
    if (N - p1[-1]) < 0.75 * dpart and (N - p1[-1]) > 0:
        p1 = np.delete(p1, -1)
        p2 = np.delete(p2, -2)
    return p1, p2


def minmed(data, nsection = None, nspec = None, npart = 1, method = 'MedMed'):
    """
    data: 2D
    npart: divide the data into n parts
    nsection: divide each part into n sections
    then calculate the median of each section.
    use the 'median' or 'min' as the baseline of this part.
    Ref: Putman et al. 2002 
    https://ui.adsabs.harvard.edu/link_gateway/2002AJ....123..873P/doi:10.1086/338088
    """
    bn = check_bottleneck()
    
    N = data.shape[0]
    print(f"Divided the data into {npart} part(s).")
    if nsection is None: 
        print(f"Each part has {N//npart//nspec} sections. Each section has {nspec} specs.")
    elif nspec is None:
        print(f"Each part has {nsection} sections. Each section has {N//npart//nsection} specs.")
    
    p1s, p2s = get_trange(N, npart)
    
    bsl = np.zeros_like(data)
    for p1, p2 in zip(p1s, p2s):
        print("part:",[p1, p2])
        s1, s2 = get_trange(p2 - p1, nsection, nspec)
        ind1, ind2 = s1 + p1, s2 + p1

        meds = np.zeros(np.hstack((len(ind1),data.shape[1:])))
        for i in tqdm(range(len(ind1))):
            meds[i] = bn.nanmedian(data[ind1[i]:ind2[i]], axis = 0)
        
        if method == 'MinMed':
            med = bn.nanmin(meds, axis = 0)
        elif method == 'MedMed':
            med = bn.nanmedian(meds, axis = 0)
            
        bsl[p1:p2] = med
        
    return bsl



#!/usr/bin/env python
# coding: utf-8

# author: Xu Chen, 2021.07
# code：Xu Chen

import numpy as np
from copy import deepcopy
from matplotlib import pyplot as plt 

def rfis_in_one_spec(spec,freq,is_rfi,mw_use,freq_step=8.1,RMS = None,freq_thr = 0.5,
                     rfi_fit_use = 'two groups',step=0,ylim = None,plot = False,
                     pdf = None,**kwargs): 

    is_rfi_ = deepcopy(is_rfi)
    is_rfi_[mw_use] = False
    syn_rfi1 = [];syn_rfi2 = [];syn_rfi3 = []
    syn_freq1 = [];syn_freq2 = [];syn_freq3 = []
    
    from markRFI import find_center,find_local_peak,get_startend
    start,end = get_startend(is_rfi_,**kwargs)
    fcenter1,fcenter2,fcenter3,fc0 = find_center(spec,freq,is_rfi_,RMS,freq_step=freq_step,
                                    freq_thr =freq_thr,rfi_fit_use = rfi_fit_use,**kwargs)
    if plot:
        
        if pdf is not None:
            plt.switch_backend('agg')

        fig2,ax = plt.subplots(figsize=(15,3))
        ax.plot(freq,spec)
        ymin = np.min(spec);ymax = np.max(spec)*.8
        ax.vlines(fc0,ymin=ymin,ymax=ymax,linestyles='--',colors='k',label = 'all')
        ax.vlines(fcenter1,ymin=ymin,ymax=ymax,linestyles='--',colors='r',label = '1')
        ax.vlines(fcenter2,ymin=ymin,ymax=ymax,linestyles='--',colors='g',label = '2')
        if (rfi_fit_use == 'three groups') & (fcenter3.size>0):
            ax.vlines(fcenter3,ymin=ymin,ymax=ymax,linestyles='--',colors='b',label = '3')
        ax.grid();ax.legend();
        if ylim is not None:
            ax.set_ylim(ylim[0],ylim[1])
        if pdf is not None:
            pdf.savefig();plt.close()

        fig1 = plt.figure(figsize=(10,4))
        ax1 = fig1.add_subplot(131)
        ax2 = fig1.add_subplot(132)
        ax3 = fig1.add_subplot(133)
        ax1.grid();ax2.grid();ax3.grid()
    
    k1,k2,k3 = 0,0,0
    
    for s,e in zip(start,end):
        rfi1 = spec[s:e]
        freq1 = freq[s:e]
        freq_peak,flux_peak = find_local_peak(rfi1,freq1,RMS=RMS,)

        vc = freq_peak
        if (vc > freq1[0])&(vc < freq1[-1]):
            norm_factor = 1/(np.sum(rfi1))
            nrfi = rfi1 * norm_factor
            nfreq = freq1 - vc

            if vc in fcenter1:
                if plot:
                    ax1.step(nfreq,nrfi+k1,label = f'{vc:.3f}')
                syn_rfi1.append(nrfi)
                syn_freq1.append(nfreq)
                k1 += step
            elif vc in fcenter2:
                if plot:
                    ax2.step(nfreq,nrfi+k2,label = f'{vc:.3f}')
                syn_rfi2.append(nrfi)
                syn_freq2.append(nfreq)
                k2 += step
            elif vc in fcenter3:
                if plot:
                    ax3.step(nfreq,nrfi+k3,label = f'{vc:.3f}')
                syn_rfi3.append(nrfi)
                syn_freq3.append(nfreq)
                k3 += step
    if plot:
        #ax.legend(loc=2,borderaxespad=1,bbox_to_anchor=(1.0,1.0))
        ax1.set_xlabel('MHz');ax2.set_xlabel('MHz');ax3.set_xlabel('MHz');
        ax1.set_ylabel('flux');ax2.set_ylabel('flux');ax3.set_ylabel('flux')
        ax1.set_title('RFI1 in one spec in a beam');
        ax2.set_title('RFI2 in one spec in a beam');
        ax3.set_title('RFI3 in one spec in a beam');
        if pdf is not None:
            pdf.savefig();plt.close()

    return (syn_rfi1,syn_rfi2,syn_rfi3),(syn_freq1,syn_freq2,syn_freq3)


def synthetic_rfi(syn_rfi,syn_freq,fdelta,plot = False,pdf = None,):
    
    neg_num = [np.sum(syn_freq[fn] < 0) for fn in range(len(syn_freq))]
    pos_num = [len(syn_freq[fn]) - neg_num[fn] for fn in range(len(syn_freq))] 
    
    m = np.median(neg_num) 
    n = np.median(pos_num) 
    x_inpd = np.arange(-fdelta*m,fdelta*n,fdelta)
    from scipy.interpolate import interp1d
    syn_rfi_inpd = np.zeros((len(syn_rfi),len(x_inpd)))
    
    if plot:
        if pdf is not None:
            plt.switch_backend('agg')
        fig,ax = plt.subplots(figsize=(5,4))
        
    for j in range(len(syn_rfi)):
        y_interp = interp1d(syn_freq[j],syn_rfi[j],kind='linear',fill_value="extrapolate")
        y_inpd = y_interp(x_inpd)

        syn_rfi_inpd[j,:] = y_inpd
        if plot:
            ax.step(x_inpd,y_inpd)

    syn_rfi_med = np.median(syn_rfi_inpd,axis = 0)
    syn_peak = np.max(syn_rfi_med)
    if plot:
        ax.plot(x_inpd,syn_rfi_med,'k',label = 'interplate median')
        ax.set_title('RFIs after interpolation');
        ax.legend()
        if pdf is not None:
            pdf.savefig();plt.close()

    return x_inpd,syn_rfi_inpd,syn_rfi_med, syn_peak, (m,n)

def common_shape(spec,ori_spec,freq,pd_rfi,theorys,syn_rfis,syn_freqs,m31_use,RMS,
                 freq_thr=.3,plot = False,pdf = None,only_M31 = False,ylim = None,**kwargs):

    fdelta = freq[1]-freq[0]
    #############
    theory1,theory2,theory3 = theorys
    syn_rfi1,syn_rfi2,syn_rfi3 = syn_rfis
    syn_freq1,syn_freq2,syn_freq3 = syn_freqs
    
    x_inpd1,_,syn_rfi_med1, syn_peak1,(m1,n1) = synthetic_rfi(syn_rfi1,syn_freq1,fdelta,plot,pdf)
    x_inpd2,_,syn_rfi_med2, syn_peak2,(m2,n2) = synthetic_rfi(syn_rfi2,syn_freq2,fdelta,plot,pdf)
    if len(syn_rfi3) > 0:
        x_inpd3,_,syn_rfi_med3, syn_peak3,(m3,n3) = synthetic_rfi(syn_rfi3,syn_freq3,fdelta,plot,pdf)

    #################
    pd_rfi_ = deepcopy(pd_rfi)
    pd_rfi_[m31_use] = False
    from markRFI import get_startend,find_local_peak
    start,end = get_startend(pd_rfi_,**kwargs)
    
    def replace_rfi(cfrq,peak,fdelta,m,n,com_rfi,x_inpd,syn_rfi_med,syn_peak):
        fl,fr = 0 - fdelta * m, 0 + fdelta * n
        if fl < freq[0] - cfrq: fl = freq[0] - cfrq 
        if fr > freq[-1] - cfrq: fr = freq[-1] - cfrq 
        rfi_use = (freq >= cfrq - fdelta * m)&(freq < cfrq + fdelta * n)
        syn_use = (x_inpd >= fl)&(x_inpd <= fr)
        syn_rfi_med_ = syn_rfi_med[syn_use]
        
        rfi_len = np.sum(rfi_use)
        syn_len = len(syn_rfi_med_)
        diflen = syn_len - rfi_len
        if diflen > 0:
            syn_rfi_med_ = syn_rfi_med_[:-int(diflen)]
        elif diflen < 0:
            rfi_use[:-int(diflen)] = False
            
        com_rfi[rfi_use] = syn_rfi_med_ * peak / syn_peak
        return com_rfi
    
    com_rfi = np.full_like(freq,np.nan)
    for s,e in zip(start,end):
        rfi1 = spec[s:e]
        freq1 = freq[s:e]
        cfrq,flux_peak = find_local_peak(rfi1,freq1,RMS=RMS,)
        peak = flux_peak + RMS
        if np.min(np.abs(cfrq - theory1)) < freq_thr:
            com_rfi = replace_rfi(cfrq,peak,fdelta,m1,n1,com_rfi,x_inpd1,syn_rfi_med1,syn_peak1)
            
        elif np.min(np.abs(cfrq - theory2)) < freq_thr:
            com_rfi = replace_rfi(cfrq,peak,fdelta,m2,n2,com_rfi,x_inpd2,syn_rfi_med2,syn_peak2)
        else:
            if theory3.size > 0 & len(syn_rfi3) > 0:
                if np.min(np.abs(cfrq - theory3)) < freq_thr:
                    com_rfi = replace_rfi(cfrq,peak,fdelta,m3,n3,com_rfi,x_inpd3,syn_rfi_med3,syn_peak3)
                    
    ret = ori_spec - com_rfi
    ret[np.isnan(ret)] = spec[np.isnan(ret)]
    
    if only_M31:
        m31_unuse = (freq<m31_frange[0]-3)|(freq>m31_frange[1]+3)
        ret[m31_unuse] = spec[m31_unuse]
    
    if plot:
        if pdf is not None:
            plt.switch_backend('agg')
        fig,ax = plt.subplots(figsize=(15,3))
        ax.plot(freq,spec,label='ori')
        ax.grid()
        if ylim is not None:
            plt.ylim(ylim[0],ylim[1])
        ax.plot(freq,com_rfi,label = 'common rfi')
        ax.plot(freq,ret,label = 'diff')
        ax.legend()

        fig,ax = plt.subplots(figsize=(15,3))
        ax.plot(freq,spec,label='ori')
        ax.grid()
        if ylim is not None:
            plt.ylim(ylim[0],ylim[1])
        ax.plot(freq,com_rfi,label = 'common rfi')
        ax.plot(freq,ret,label = 'diff')
        ax.legend();plt.xlim(1400,1440)
        if pdf is not None:
            pdf.savefig();plt.close()
    
    return ret,com_rfi





__all__ = ['plot', 'colors', 'linestyles', 'linewidths']


import h5py
import os
import re
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt


# for 19 beams
colors = ['k'] + ['C0', 'C1', 'C2', 'C3', 'C4', 'C5']*3
linestyles = ['-'] +  ['-']*6 + ['-.']*6 + ['--']*6
linewidths = [3] + [1]*6 + [2]*6 + [2]*6

def plot(info, frange=[1050, 1450], suptitle='', K_Jy_ratio=False, field_start=None):

    fig, axs = plt.subplots(2, 1, figsize=(10, 14), sharex=True, sharey=True)
    f = info
    freq = f['freq'][:]
    is_use = (freq > frange[0]) & (freq < frange[1])
    freq = freq[is_use]
    keys = []
    if field_start is None:
        for key in f.keys():
            if key.startswith('M'):
                keys.append(key)
        if len(keys) == 0:
            for key in f.keys():
                if key.startswith('K_Jy'):
                    keys.append(key)
    else:
        for key in f.keys():
            if key.startswith(field_start):
                keys.append(key)
    if len(key) > 19:
        raise(ValueError('find more than 19beams?'))
    # plot
    ax = axs[0]
    polar = 0

    if K_Jy_ratio:
        refer_d = {}

    for i, key in enumerate(keys):
        y = f[key][0, :, polar][is_use]
        label = key
        if K_Jy_ratio:
            nB = int(re.sub(r'\D', '', key))
            print(nB)
            ra = f[f'ra{nB}'][:1]
            dec = f[f'dec{nB}'][:1]
            mjd = f[f'mjd'][:1]

            refer_y = gain.Get_gain(ra, dec, mjd, nB, freq)[0][0]
            refer_d[nB] = refer_y
            y = y/refer_y
            label += '_r'
        ax.plot(freq, y, c=colors[i], ls=linestyles[i], lw=linewidths[i], label=label)
    ax.grid()
    ax.set_xlabel('freq')
    ax.set_title(f'Polar:{polar}')
    ## plot
    if f[key].shape[2] == 2:
        polar = 1
        ax = axs[1]

        for i, key in enumerate(keys):
            y = f[key][0, :, polar][is_use]
            label = key
            if K_Jy_ratio:
                nB = int(re.sub(r'\D', '', key))
                refer_y = refer_d[nB]
                y = y/refer_y
                label += '_r'
            ax.plot(freq, y, c=colors[i], ls=linestyles[i], lw=linewidths[i], label=label)
    ax.grid()
    ax.set_title(f'polar:{polar}')
    axs[1].legend(loc=(1.01, 0))
    axs[0].legend(loc=(1.01, 0))

    fig.suptitle(suptitle)
    fig.tight_layout()
    # for ax in axs:
    #     ax.set_ylim(13, 17.9)

    return fig


if __name__ == '__main__':
    import argparse
    import os
    import re

    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('fpaths', nargs='+',
                        help='paths of FluxGain files')
    parser.add_argument('--outdir', default='./',
                       help='default is ./')
    args = parser.parse_args()
    plot_files(args.fpaths, args.outdir)





__all__ = ['gen_radec_file', 'get_radec']


import os
import re
import sys
from .utils.io import get_nB, replace_nB


def gen_radec_file(file_spec, paras=[]):
    """
    generating ra dec for a specs_T file
    """
    import subprocess
    command = [sys.executable, '-m', 'hifast.radec', file_spec] + paras
    print('run:')
    print(' '.join(command))
    sys.stdout.flush()
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # process.wait()
    print(*process.communicate())
    if process.returncode != 0:
        raise(ValueError(f'fail to generate the radec file of {file_spec}'))


import h5py

def get_radec(file_spec, mjd, nB_radec=1, file_radec = None):
    """
    load the ra dec of spectra in file_spec (*specs_T.hdf5)

    file_spec: str; specs fpath, used to guess radec filename
    mjd: np.array; used to check

    """
    nB = get_nB(file_spec)
    if file_radec is None:
        # ra dec file
        file_radec = file_spec.rsplit('specs_T', 1)[0] + 'specs_T-radec.hdf5'
        #file_radec = re.findall(r'.*-specs_T', file_radec)+'-radec.hdf5'
        file_radec_ori = file_radec
        # default is in beam `nB_radec`
        file_radec = replace_nB(file_radec, nB_radec)

    if os.path.exists(file_radec):
        f = h5py.File(file_radec, 'r')
    elif os.path.exists(file_radec_ori):
        f = h5py.File(file_radec_ori, 'r')
    else:
        raise(OSError(f"can not find the RA DEC file \n{file_radec}\n, please generate it by using 'python -m hifast.radec'"))

    # check if mjd match
    mjd_match = False if len(mjd) != len(
        f['S']['mjd'][:]) else (mjd == f['S']['mjd'][:]).all()
    if not mjd_match:
        print('the mjd of spec file is not same with that in ', file_radec)
        sys.stdout.flush()
        file_radec_2 = replace_nB(file_radec, nB)
        if not os.path.exists(file_radec_2):
            print('try to generate ', file_radec_2)
            sys.stdout.flush()
            header = f['Header'].attrs
            import json
            for key in header.keys():
                argv = json.loads(header[key])['argv']
                paras = argv.split()
                if os.path.basename(paras[0]) not in ['cli_radec.py', 'radec.py']:
                    continue
                else:
                    paras.pop(0)
                    for s in paras:
                        if '.hdf5' in s:
                            paras.remove(s)
            gen_radec_file(file_spec, paras)
            sys.stdout.flush()
        f.close()
        f = h5py.File(file_radec_2, 'r')
        print('using', file_radec_2)
        sys.stdout.flush()
    # check again
    mjd_match = False if len(mjd) != len(
        f['S']['mjd'][:]) else (mjd == f['S']['mjd'][:]).all()
    if not mjd_match:
        raise(ValueError(
            f'the mjd of spec file is not same with that in {file_radec_2} Abort...'))
    # load
    ra = f['S']['ra'+'%d' % nB][:]
    dec = f['S']['dec'+'%d' % nB][:]
    if 'is_extrapo' in f['S'].keys():
        is_extrapo = f['S']['is_extrapo'][:]
    else:
        is_extrapo = None
    f.close()
    return ra, dec, is_extrapo


if __name__ == '__main__':
    import h5py
    import argparse
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('fname',
                        help='file name')
#     parser.add_argument('-f', '--force', action='store_true',
#                         help='overwriting file if out file exists')
    parser.add_argument('--nB_radec', type=int, default=1,
                        help='Beam number of the radec file name')
    parser.add_argument('--file_radec',
                        help='file radec name if it exists')
#     parser.add_argument('--outdir',
#                         help='default is same with the input file')
    args = parser.parse_args()
    f = h5py.File(args.fname, 'r+')
    g = f['S']

    values = get_radec(args.fname, g['mjd'][:], args.nB_radec, args.file_radec)

    keys = ['ra', 'dec', 'is_extrapo']

    for key, value in zip(keys, values):
        if value is None:
            continue
        if key in g.keys():
            g[key][:] = value
        else:
            g[key] = value
    f.close()





__all__ = ['ConvFun']


import numpy as np
from scipy import special
from functools import partial


class ConvFun(object):
    def __init__(self, kernel_type, beam_fwhw, *,
                gaussian_fwhw=None,
                bsize=None, gsize=None,
                wei_min=None):
        """
        kernel_type: str
            'gaussian':
            'bessel_gaussian':
            'sinc_gaussian':

        """
        if kernel_type == 'gaussian':
            if gaussian_fwhw is None:
                gaussian_fwhw = beam_fwhw/2
            self.sigma = gaussian_fwhw / (2*(2*np.log(2))**0.5)
            self.dis_cut_suggest = 3*self.sigma
            self.gaussian_fwhw = gaussian_fwhw

        elif kernel_type == 'bessel_gaussian' or kernel_type == 'sinc_gaussian':
            self.bsize = 1.55 * beam_fwhw / 3. if bsize is None else bsize
            self.gsize = 2.52 * beam_fwhw / 3. if gsize is None else gsize
            if kernel_type == 'bessel_gaussian':
                # first zero special.jn_zeros(1, 1) = 3.8317059702075
                self.dis_v0 = 3.8317059702075*self.bsize/np.pi
            elif kernel_type == 'sinc_gaussian':
                self.dis_v0 = self.bsize
            self.dis_cut_suggest = self.dis_v0
        else:
            raise(ValueError(f'kernel \"{kernel_type}\" is not supported'))

        self.kernel_type = kernel_type
        self.wei_min = wei_min
        self.beam_fwhw = beam_fwhw

        self.beam_change()

    def beam_change(self):
        # calculate new beam fwhw
        if self.kernel_type == 'gaussian':
            self.new_beam_fwhw = (self.beam_fwhw**2 + self.gaussian_fwhw**2)**0.5
            # volume under the Gaussian function: 2*pi*A*sigma_X*sigma_Y
            self.beam_factor = self.new_beam_fwhw**2 / self.beam_fwhw**2

        else:
            # To do: use Numerical method to calculate
            pass

    def __call__(self, dis):

        wei = getattr(self, f"{self.kernel_type}_kernel")(dis)

        if self.wei_min is not None:
            wei[wei < self.wei_min] = 0

        return wei

    def gaussian_kernel(self, dis):
        sigma = self.sigma
        w = np.exp(-dis**2/(2*sigma**2))
        return w

    def bessel_gaussian_kernel(self, dis):
        """
        """
        a, b = self.bsize, self.gsize
        x = np.pi*dis/a
        wei = np.empty_like(dis)
        not_zero = dis != 0
        wei[~not_zero] = 1 # peak value depend on the not_zero calculating method
        wei[not_zero] = 2*special.j1(x[not_zero])/(x[not_zero])*np.exp(-(dis[not_zero]/b)**2)

        wei[abs(dis) > self.dis_v0] = 0
        return wei

    def sinc_gaussian_kernel(self, dis):

        a, b = self.bsize, self.gsize
        x = np.pi*dis/a
        wei = np.empty_like(dis)
        not_zero = dis != 0
        wei[~not_zero] = 1
        wei[not_zero] = np.sin(x[not_zero])/(x[not_zero])*np.exp(-(dis[not_zero]/b)**2)

        wei[abs(dis) > self.dis_v0] = 0
        return wei





__all__ = ['IO', 'IO_i']


from .utils.io import *
from copy import deepcopy


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='Fit and subtract standing wave',)
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input baselined spectra file path')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
parser.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                    help="if set, don't check or add ra dec")
parser.add_argument('--nproc', '-n', type=int, default=1,
                    help='number of process used in fitting baseline')
parser.add_argument('--show_prog', type=bool_fun, choices=[True, False], default='True', env_var='HIFAST_SHOW_PROG',
                    help='show progress bar')

# smooth
group = parser.add_argument_group('For FFT, smooth to find where should be replaced; for sin-fitting, preprocess.')
group.add_argument('--s_method_t', default='gaussian', choices=['none', 'gaussian', 'boxcar', 'median'],
                   help='smooth method along time axis')
group.add_argument('--s_sigma_t', type=int, default=5,
                   help='smooth sigma along time axis, size = (2*sigma+1) for boxcar and median')
group.add_argument('--s_method_freq', default='gaussian', choices=['none', 'median', 'gaussian', 'boxcar'],
                   help='smooth method along freq axis')
group.add_argument('--s_sigma_freq', type=float, default=3,
                   help='smooth sigma along freq axis, size = (2*sigma+1) for boxcar and median')

group = parser.add_argument_group('For FFT, smooth for finding Troughs of standing waves')
group.add_argument('--s_method_t_T', default='gaussian', choices=['none', 'gaussian', 'boxcar', 'median'],
                   help='smooth method along time axis')
group.add_argument('--s_sigma_t_T', type=int, default=10,
                   help='smooth sigma along time axis, size = (2*sigma+1) for boxcar and median')
group.add_argument('--s_method_freq_T', default='gaussian', choices=['none', 'median', 'gaussian', 'boxcar'],
                   help='smooth method along freq axis')
group.add_argument('--s_sigma_freq_T', type=float, default=5,
                   help='smooth sigma along freq axis, size = (2*sigma+1) for boxcar and median')

# method
group = parser.add_argument_group(f'*select method used to fit standing wave\n{sep_line}')
group.add_argument('--method', default='fft', choices=['sin_poly', 'fft', 'running_median', 'running_mean'],
                   help='method to fit standing wave')
group.add_argument('--nobld', type=bool_fun, choices=[True, False], default='False',
                   help="if True, use the spectra before bld to subtract standing wave and output file name will add 'nobld'")
group.add_argument('--fpattern_nobld',
                   help='if not specify, guess from the History recored in the fpath')

# least square
group = parser.add_argument_group(f'*parameters for --method sin_poly\n{sep_line}:' +
                                  '\n'+'addition options for preprocessing before fitting')
group.add_argument('--njoin', type=int, default=0,
                   help='join nspec along t')
group = parser.add_argument_group('sinusoidal standing wave (sw) fitting')
group.add_argument('--exclude_m', type=int, default=0,
                   help='')
group.add_argument('--sin_f', type=float, nargs='+', default=0.929,
                   help='Initial guess for sw freq')
group.add_argument('--bound_f', type=float, nargs=2, default=[.90, .95],
                   help='Lower and upper bounds for sw freq')
group.add_argument('--deg', type=int, default=1,
                   help='Degree of the added polynomial')
# fft
group = parser.add_argument_group(f'*parameters for --method fft\n{sep_line}:' +
                                  '\n'+'replace big RFI or sources firstly')
group.add_argument('--iter_twice', type=bool_fun, choices=[True, False], default='True',
                   help='iterate twice? first remove the ripples and then back to find replace area again.')
# replace big RFI or sources firstly
group.add_argument('--rfi_method', default='near_ripple',
                   choices=['near_ripple','zero_ripple',],
                   help='method to replace big RFI, recommend the first one')
group.add_argument('--mw_frange', type=float, nargs=2, default=[0, 0],
                   help='milky way freq range. If do not use, keep it as default.')
group.add_argument('--rms_sigma', type=float, default=6,
                   help='gauss filter sigma to compute real rms')
group.add_argument('--rms_frange', type=float, nargs=2, default=[0, 0],
                   help='freq range to compute rms')
group.add_argument('--rms_step', type=float, default=5,
                   help='a step (MHz) to find where to compute real rms')

group.add_argument('--times_thr', type=float, default=4,
                   help='sparks above ~ times of rms will be set noise (unsmooth)')
group.add_argument('--times_s_thr', type=float, default=3,
                   help='above ~ times of rms will be replaced (smoothed once)')
group.add_argument('--times_s_thr2', type=float, default=2,
                   help='above ~ times of rms will be replaced (smoothed twice)')
group.add_argument('--ext_freq', type=float, default=1.3,
                   help='extend freq range to replace (mhz)')
group.add_argument('--rfi_width_lim', type=float, default=20,
                   help='rfi should contain more channels than limit')
group.add_argument('--ext_sec', type=int, default=20,
                   help='extend channel number of start and end of each section')
group.add_argument('--restrict_bound', type=bool_fun, choices=[True, False], default='False',
                   help='restrict replace area bound')
group.add_argument('--save_is_excluded', type=bool_fun, choices=[True, False], default='False',
                   help='save the replaced area as is_excluded')

group = parser.add_argument_group(f'phase space')
# remove which component in fft? (only 4 types now)
group.add_argument('--sw_base', type=bool_fun, choices=[True, False], default='True',
                   help='if True, remove constant components')
group.add_argument('--sw_periods', nargs='+', choices=['1mhz', '2mhz', '0_04mhz', 'none'], default=['1mhz',  '0_04mhz'],
                   help='remove ripple (1mhz: 1.08mhz, 2mhz:1.92mhz, 0_04mhz: 0.039 mhz)')
group.add_argument('--check_2mhz', type=bool_fun, choices=[True, False], default='False',
                   help='if True, remove 2mhz from sw_periods except for Beam 6')

group.add_argument('--amp_thr_mean_factor', type=float, nargs=2, default=[1.05, 1.3],
                   help='above mean amptitude threshold will be chosed, noise off and on')
group.add_argument('--amp_thr_solo_factor', type=float, nargs=2, default=[1.4, 1.7],
                   help='above amptitude threshold will be chosed in every spec, noise off and on')
group.add_argument('--chan_wide', type=int, default=5,
                   help='channel numbers near 1mhz to be chosed (wide)')
group.add_argument('--chan_narr', type=int, default=2,
                   help='channel numbers near 1mhz to be chosed (narrow)')
group.add_argument('--choose_method', default='all', choices=['all', 'interpolate'],
                   help='method to choose components in fft')

# running median or mean
group = parser.add_argument_group(f'*parameters for --method running_median or running_mmean\n{sep_line}')
group.add_argument('--nspec',type=int, default=200,
                    help='average how many specs to fit baseline.')
group.add_argument('--func', default='iter', choices=['iter', 'smooth'],
                       help='function to mean or median')

# Exclude files
group = parser.add_argument_group(f'Exclude known source catalogs\n{sep_line}')
group.add_argument('--src_file',
                   help='Path to a text catalog with columns for RA [deg], Dec [deg], radius [arcmin], minimum frequency [MHz], and maximum frequency [MHz]. Identified sources will be added to the is_excluded array.')
group.add_argument('--frame', choices=['BARYCENT', 'HELIOCEN', 'LSRK', 'LSRD'], default='LSRK',
                   help='Specify the velocity rest frame for excluding source frequency ranges.')

# interaction
group = parser.add_argument_group(f'*Interaction\n{sep_line}')
group.add_argument('-i', '--interact', action='store_true', not_in_write_out_config_file=True,
                   help='interaction')
group.add_argument('--ylim', nargs='+', default=['auto'], not_in_write_out_config_file=True,
                   help='ylim')
group.add_argument('--figsize', type=float, nargs=2, default=(10, 7), not_in_write_out_config_file=True,
                   help='figsize')
group.add_argument('--length', type=int, default=40, not_in_write_out_config_file=True,
                   help='spetra numbers used to test')
group.add_argument('--start_init', type=int, default=0, not_in_write_out_config_file=True,
                   help='the index of the spectrum shown at start')


class IO(BaseIO):
    ver = 'old'

    def __init__(self, *args, dict_bef_bld=None, **kwargs):
        self.dict_bef_bld = dict_bef_bld
        super().__init__(*args, **kwargs)

    def _get_fpart(self,):
        if self.args.nobld:
            return '-sw_nobld'
        return '-sw'

    def _import_m(self,):
        """
        need modify this function
        """
        super()._import_m()
        global h5py, np, OrderedDict, sub_baseline, get_exclude_fun
        import h5py
        import numpy as np
        from collections import OrderedDict
        from .core.baseline import sub_baseline, get_exclude_fun

    @staticmethod
    def find_fname_his(Header, fname):
        """
        from historys in Header to find file:
        return: [path, path related to the 'outdir' in history]
        """
        import json
        fpath_in_list = []
        outdir_list = []
        cwd_list = []
        for key in Header.keys():
            if key.startswith('HISTORY'):
                his = json.loads(Header[key])
                try:
                    args_h = json.loads(his['args'])
                    fpath_in_list += [args_h['fpath']]
                    outdir_list += [args_h['outdir']]
                    cwd_list += [his['cwd']]
                except:
                    try:
                        import re
                        fpath_in_list += [re.findall(r'fpath=.*.hdf5', his['args'])[0].split('=')[1][1:]]
                        outdir_list += [re.findall(r'outdir=[^,]*', his['args'])[0].split('=')[1][1:-1]]
                        cwd_list += [his['cwd']]
                    except:
                        pass
        for i, val in enumerate(fpath_in_list):
            if os.path.basename(val) == fname:
                path_ = os.path.abspath(os.path.join(cwd_list[i], val))
                out = [path_, os.path.relpath(path_, os.path.join(cwd_list[i], outdir_list[i]))]
                break
        return out

    def _gen_s2p_ori_fpath(self,):
        """
        find file path of spectra before baseline subtracted
        """
        args = self.args
        if args.fpattern_nobld is not None:
            print('using the path from --fpattern_nobld as the spectra file before baseline subtracted')
            ori_fpath = self.replace_nB(args.fpattern_nobld, self.nB)
            if not os.path.exists(ori_fpath):
                raise(FileNotFoundError(f'file {ori_fpath} not exists'))

        else:
            print('guess the path of the spectra file before baseline subtracted from the History')
            fname_find = os.path.basename(args.fpath).rsplit('bld')[0][:-1] + '.hdf5'
            ori_fpaths = self.find_fname_his(self.Header, fname_find)
            if os.path.exists(ori_fpaths[0]):
                ori_fpath = ori_fpaths[0]
            elif os.path.exists(ori_fpaths[1]):
                ori_fpath = ori_fpaths[1]
            else:
                raise(FileNotFoundError("can't find the spectra file before baseline subtracted, \
                                        please specify --fpattern_nobld"))
        self.s2p_ori_fpath = ori_fpath
        self.Header['ori_fpath'] = self.s2p_ori_fpath

    def _load_s2p_ori(self):
        """
        load spectra before baseline subtracted
        """
        if self.dict_in is None:
            self._gen_s2p_ori_fpath()
            print(f'load the spectra before baseline subtracted from: \n{self.s2p_ori_fpath}')
            f = h5py.File(self.s2p_ori_fpath, 'r')['S']
        else:
            if self.dict_bef_bld is None:
                raise(ValueError('need input dict_bef_bld'))
            f = self.dict_bef_bld
        freq = f['freq'][:]
        s2p_ori = f[self.outfield][:]
        is_use_freq = (freq >= self.freq.min()) & (freq <= self.freq.max())
        if not np.all(is_use_freq):
            inds = np.where(is_use_freq)[0]
            s2p_ori = s2p_ori[..., inds[0]:inds[-1]+1]  # freq axis at end
        s2p_ori = PolarMjdChan_to_MjdChanPolar(s2p_ori)
        return s2p_ori


    def _load_sources(self):
        args = self.args
        fpath = args.src_file
        if fpath is not None:
            from .core.regions import mask_srcs
            is_excluded = self.is_excluded if hasattr(self,'is_excluded') else np.zeros(self.s2p.shape[:2], dtype = bool)
            print(f"loading excluded files from {fpath}")
            self.is_excluded = mask_srcs(fpath, is_excluded, self.ra, self.dec, self.freq,
                                         self.mjd, inplace=True, rest_frame=args.frame)


    def _load_is_rfi(self):
        # load is_rfi, is_on
        dict_ = self.fs if self.dict_in is None else self.dict_in
        is_rfi = dict_['is_rfi'][:] if 'is_rfi' in dict_.keys() else None
        if is_rfi is not None and self.is_use_freq is not None:
            is_rfi = is_rfi[:, self.is_use_freq]
        self.is_rfi = is_rfi
        self.is_on = dict_['is_on'][:]


    def fit_sw(self, s2p, freq, args, subtract):
        """
        sin poly
        """
        fit_kwargs = {}
        keys = ['method',
                'njoin', 's_method_t', 's_sigma_t', 's_method_freq', 's_sigma_freq',
                'deg', 'sin_f']
        for key in keys:
            fit_kwargs[key] = getattr(args, key)
        fit_kwargs['nproc'] = args.nproc
        fit_kwargs['rew'] = False
        fit_kwargs['niter'] = 1
        if args.method == 'sin_poly':
            # sin_poly: optimize.minimize
            bounds = [(0., 1.), args.bound_f, (0, 2*np.pi), (-1, 1)] + [(-np.inf, np.inf), ]*args.deg
        opt_para = {'bounds': bounds, }
        fit_kwargs['exclude_fun'] = get_exclude_fun(args.exclude_m)
        m, f = s2p.shape[0], s2p.shape[1]
        fit_kwargs['is_excluded'] = np.full((2, m, f), self.is_excluded).transpose(1, 2, 0) if hasattr(self, 'is_excluded') else None
        return sub_baseline(freq, s2p, subtract=subtract, opt_para=opt_para, **fit_kwargs)

    def check_rms_range(self, is_rfi, rms_step=5):
        args = self.args
        rms_frange = args.rms_frange
        if (rms_frange is None) or (rms_frange[1] - rms_frange[0] <= 0):
            from .ripple.markRFI import get_rms_frange
            if is_rfi is not None:
                is_lf = np.all(is_rfi, axis = 1)
                is_excluded = np.any(is_rfi[~is_lf], axis = 0)
                data = self.s2p[~is_lf]
            else:
                data = self.s2p
                is_excluded = None
            spec = np.nanmean(np.nanmean(data, axis = 0),axis = -1)
            self.args.rms_frange = get_rms_frange(spec,self.freq,rms_step=5,is_excluded = is_excluded)
        else:
            freq = self.freq
            if (rms_frange[0] < freq[0]) or (rms_frange[1] > freq[-1]):
                raise ValueError(f"rms frange {rms_frange} is not in freq range {[freq[0],freq[-1]]}.")

    def fft_fit_sw(self, s1p, is_rfi=None, is_on=None,iter_twice = False,
                  sm_find = None, sm_trough = None, sm_res = None, ):
        """
        fft
        """
        from .ripple import sw_fft
        args = self.args

        # replace rfi and mw
        # replace args
        rep_args = {}
        self.check_rms_range(is_rfi, rms_step = args.rms_step)
        if args.rfi_method in ['near_ripple','zero_ripple',]:
            keys = ['rms_sigma', 'rms_frange', 'times_s_thr','times_s_thr2',
                    'times_thr', 'rfi_width_lim', 'ext_sec', 'ext_freq', 'mw_frange',]
        else:
            raise(ValueError('not supported rfi_method'))
        for key in keys:
            rep_args[key] = getattr(args, key)

        if not iter_twice: rep_args['times_s_thr2'] = None
        if args.rfi_method == 'zero_ripple':
            rep_args['fill'] = 'zero'
            args.rfi_method = 'near_ripple'
        rep_args['verbose'] = args.show_prog

        save_is_excluded = args.save_is_excluded & iter_twice
#         print("############ save_is_excluded", save_is_excluded)
        s1p, is_excluded = sw_fft.replace_rfi(s1p, self.freq, time_rfi=is_rfi, method=args.rfi_method,
                                  data_find = sm_find, data_trough = sm_trough, data_restrict = sm_res,
                                 save_is_excluded = save_is_excluded, **rep_args)
        # fft args
        fft_args = {}
        keys = ['chan_wide', 'chan_narr',  'choose_method',
                'choose_method', 'sw_periods', 'sw_base']
        for key in keys:
            fft_args[key] = getattr(args, key)
        if args.check_2mhz:
            print('checking 2mhz')
            try:
                fft_args['sw_periods'].remove('2mhz')
            except:
                pass
            if self.nB in [6, ]:
                fft_args['sw_periods'].insert(-1, '2mhz')
            else:
                print('beam number !=6, remove 2mhz in sw_periods if it exists')

        # use the first factor for noise off or not defined
        for key in ['amp_thr_mean_factor', 'amp_thr_solo_factor']:
            fft_args[key] = getattr(args, key)[0]

        if is_on is not None:
            ret = np.zeros_like(s1p)
            print("# noise off")
            fft_args['is_excluded_mean'] = np.all(is_rfi[~is_on], axis=1) if is_rfi is not None else None
            ret[~is_on] = sw_fft.fit_sw_fft(s1p[~is_on], self.freq, args.nproc, **fft_args)

            print("# noise on")
            for key in ['amp_thr_mean_factor', 'amp_thr_solo_factor']:
                fft_args[key] = getattr(args, key)[1]
            fft_args['is_excluded_mean'] = np.all(is_rfi[is_on], axis=1) if is_rfi is not None else None
            ret[is_on] = sw_fft.fit_sw_fft(s1p[is_on], self.freq, args.nproc, **fft_args)
        else:
            fft_args['is_excluded_mean'] = np.all(is_rfi, axis=1) if is_rfi is not None else None
            ret = sw_fft.fit_sw_fft(s1p, self.freq, args.nproc, **fft_args)
        return ret, is_excluded

    def med_fit_sw(self, s1p,):
        """
        running median or mean
        """
        from .ripple.running_median import running_median
        args = self.args

        fit_args = {}
        keys = ['nspec','func', 'nproc']
        for key in keys:
            fit_args[key] = getattr(args, key)
        fit_args['method'] = args.method[8:]

        if hasattr(self, 'is_excluded'):
            s1p[self.is_excluded | self.is_rfi] = np.nan

#         sw_on = running_median(s1p[self.is_on], **fit_args)
#         sw_off = running_median(s1p[~self.is_on], **fit_args)

#         sw = np.zeros_like(s1p)
#         sw[self.is_on] = sw_on
#         sw[~self.is_on] = sw_off
        sw = running_median(s1p, **fit_args)

        return sw


    def gen_s2p_out(self,):
        args = self.args
        # gen self.s2p_out
        s2p = self.s2p[:]

        # src add to is_excluded
        self._load_sources()
        self._load_is_rfi()

        # fit baseline:
        print(f'standing wave fitting and substract by {args.method}')
        if args.method in ['sin_poly', ]:
            if args.nobld:
                self.s2p_out = self._load_s2p_ori()[:] - self.fit_sw(s2p, self.freq, args, subtract=False)
            else:
                self.s2p_out = self.fit_sw(s2p, self.freq, args, subtract=True)
        else:
            is_rfi = self.is_rfi
            is_on = self.is_on

            s2p_out = deepcopy(s2p)

            if args.method == 'fft':
                if args.iter_twice:
                    s2p_in = s2p
                else:
                    s2p_in = self._load_s2p_ori()[:] if args.nobld else s2p

                # smooth to find areas that need replacement
                find_kwargs = {}
                keys = ['s_method_t', 's_sigma_t', 's_method_freq', 's_sigma_freq',]
                for key in keys:
                    find_kwargs[key] = getattr(args, key)
                find_kwargs['is_rfi'] = is_rfi
                find_kwargs['is_on'] = is_on
                from .ripple.util import do_smooth_onoff
                sm_find = do_smooth_onoff(s2p, **find_kwargs)
                print("------------------")

                # smooth to find troughes
                if args.rfi_method != 'zero_ripple':
                    trough_kwargs = {}
                    keys = ['s_method_t_T', 's_sigma_t_T', 's_method_freq_T', 's_sigma_freq_T',]
                    for key in keys:
                        trough_kwargs[key[:-2]] = getattr(args, key)
                    trough_kwargs['is_rfi'] = is_rfi
                    trough_kwargs['is_on'] = is_on
                    if trough_kwargs == find_kwargs:
                        sm_trough = sm_find
                    else:
                        sm_trough = do_smooth_onoff(s2p, **trough_kwargs)
                        print("------------------")
                else:
                    sm_trough = sm_find
                    trough_kwargs = {}
                    trough_kwargs['s_method_t'] = 'none'

                # smooth to restrict bounds
                if trough_kwargs['s_method_t'] != 'none' and args.restrict_bound:
                    res_kwargs = {}
                    res_kwargs['s_method_freq'] = 'gaussian'
                    res_kwargs['s_sigma_freq'] = args.rms_sigma
                    sm_res = do_smooth_onoff(s2p, **res_kwargs)
                    print("------------------")
                else:
                    sm_res = np.array([None, None])[None,None,:]

                for i in range(s2p.shape[2]):
                    s2p_out[..., i] = s2p_in[..., i] - self.fft_fit_sw(s2p[..., i], is_rfi, is_on,
                          sm_find = sm_find[..., i], sm_trough = sm_trough[..., i], sm_res = sm_res[..., i])[0]

                if args.iter_twice:
                    print("Second iter ...")
                    # smooth
                    s2p_in = self._load_s2p_ori()[:] if args.nobld else s2p
                    sm_find = do_smooth_onoff(s2p_out, **find_kwargs)
                    print("------------------")
                    if trough_kwargs['s_method_t'] != 'none' and args.restrict_bound:
                        sm_res = do_smooth_onoff(s2p_out, **res_kwargs)
                        print("------------------")
                    else:
                        sm_res = np.array([None, None])[None,None,:]

                    is_excluded = self.is_excluded if hasattr(self,'is_excluded') else np.zeros_like(s2p,dtype=bool)
                    sw2 = deepcopy(s2p)
                    for i in range(s2p.shape[2]):
                        sw2[..., i], is_excluded[..., i] = self.fft_fit_sw(s2p[..., i], is_rfi, is_on, iter_twice = True,
                               sm_find = sm_find[..., i], sm_trough = sm_trough[..., i], sm_res = sm_res[..., i])
                        s2p_out[..., i] = s2p_in[..., i] - sw2[..., i]

                    if args.save_is_excluded:
                        self.is_excluded = np.any(is_excluded,axis = -1)
                        if is_rfi is not None:
                            self.is_excluded |= is_rfi


            elif (args.method == 'running_median') or (args.method == 'running_mean'):
                s2p_out = s2p - self.med_fit_sw(s2p_out, )

            self.s2p_out = s2p_out


    def __call__(self, save=True):
        self.gen_s2p_out()
        if hasattr(self,'is_excluded'):
            print("save is_excluded :D")
            self.gen_dict_out(is_excluded = self.is_excluded)
        else:
            self.gen_dict_out()
        # save to hdf5 file
        if save:
            self.save()


def check_backend():
    import matplotlib as mpl
    if 'ipympl' not in mpl.get_backend():
        print('Please run interaction in Jupyert and \'%matplotlib ipympl\' in the notebook cell ')
        sys.exit()

interact_save_lim = 20
def interact(args):
    check_backend()
    import h5py
    from .interaction import sw_i as interact
    f = h5py.File(args.fpath, 'r')
    fs = f['S']
    if 'T' in fs.keys():
        T = fs['T']
    elif 'Ta' in fs.keys():
        T = fs['Ta']
    elif 'flux' in fs.keys():
        T = fs['flux']
    interact.T2p = T
    interact.freq = fs['freq'][:]
    interact.frange = args.frange
    interact.nproc = args.nproc
    interact.length = min(args.length, interact.T2p.shape[1])
    interact.figsize = args.figsize
    interact.ylim = args.ylim[0] if len(args.ylim) == 1 else args.ylim
    interact.start_init = args.start_init
    if T.shape[1] <= interact_save_lim:
        interact.save = True
    else:
        interact.save = False
    return interact.main()


class IO_i(IO):
    def gen_s2p_out(self,):
        args = self.args
        if args.interact:
            print('Please ensure you have adjusted the parameters for each `polar`, otherwise you will get nan vaules in the file')
            self.s2p_out = interact_spec.bld_out
            return


if __name__ == '__main__':
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    # print(parser.format_values())  # useful for logging where different settings came from

    if args_.interact:
        interact_spec, widgets = interact(args_)[0:2]
        save = IO_i(args_, HistoryAdd={'interact':str(widgets)})
        if interact_spec.save:
            print('Please run \'save()\' in the notebook cell to save your results')
    else:
        print('#'*35+'Args'+'#'*35)
        args_from = parser.format_values()
        print(args_from)
        print('#'*35+'####'+'#'*35)

        HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
        io = IO(args_, HistoryAdd=HistoryAdd)
        io()
